import sys
import time

from orangecontrib.shadow4.widgets.gui.ow_electron_beam import OWElectronBeam
from orangecontrib.shadow4.widgets.gui.plots import plot_data1D
from orangecontrib.shadow4.util.shadow4_objects import ShadowData

from oasys.widgets import gui as oasysgui

from oasys.util.oasys_util import EmittingStream

from syned.beamline.beamline import Beamline
from syned.storage_ring.magnetic_structures.insertion_device import InsertionDevice
from syned.widget.widget_decorator import WidgetDecorator

from orangewidget import gui as orangegui
from orangewidget.settings import Setting


from shadow4.sources.wiggler.s4_wiggler import S4Wiggler
from shadow4.sources.wiggler.s4_wiggler_light_source import S4WigglerLightSource
from shadow4.beamline.s4_beamline import S4Beamline

class OWWiggler(OWElectronBeam, WidgetDecorator):

    name = "Wiggler Light Source"
    description = "Wiggler Light Source"
    icon = "icons/wiggler.png"
    priority = 3


    inputs = []
    WidgetDecorator.append_syned_input_data(inputs)

    outputs = [{"name":"Shadow Data",
                "type":ShadowData,
                "doc":"",}]

    magnetic_field_source = Setting(1)
    number_of_periods = Setting(1)
    k_value = Setting(10.0)
    id_period = Setting(0.010)
    file_with_b_vs_y = Setting("https://raw.githubusercontent.com/srio/shadow4/master/examples/sources/BM_multi.b")
    file_with_harmonics = Setting("tmp.h")

    shift_x_flag = Setting(0)
    shift_x_value =Setting(0.0)

    shift_betax_flag = Setting(0)
    shift_betax_value = Setting(0.0)

    e_min = Setting(0.4)
    e_max = Setting(0.4)
    number_of_rays = Setting(500)
    seed = Setting(5676561)

    # number of points in advanced parameters
    ng_e = Setting(101)
    ng_j = Setting(501)

    epsi_dx = Setting(0.0)
    epsi_dz = Setting(0.0)

    plot_wiggler_graph = 1

    beam_out = None


    def __init__(self):
        super().__init__()

        tab_wiggler = oasysgui.createTabPage(self.tabs_control_area, "Wiggler Setting")

        # wiggler parameters box
        left_box_3 = oasysgui.widgetBox(tab_wiggler, "Wiggler Parameters", addSpace=False, orientation="vertical", height=200)

        orangegui.comboBox(left_box_3, self, "magnetic_field_source", tooltip="magnetic_field_source", label="Type", items=["conventional/sinusoidal", "B from file (y [m], Bz [T])", "B from harmonics"], callback=self.set_visibility, labelWidth=220, orientation="horizontal")

        oasysgui.lineEdit(left_box_3, self, "number_of_periods", "Number of Periods", tooltip="number_of_periods",labelWidth=260, valueType=int, orientation="horizontal")

        self.conventional_sinusoidal_box = oasysgui.widgetBox(left_box_3, "", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.conventional_sinusoidal_box, self, "k_value", "K value", labelWidth=260, tooltip="k_value", valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.conventional_sinusoidal_box, self, "id_period", "ID period [m]", labelWidth=260, tooltip="id_period", valueType=float, orientation="horizontal")

        self.b_from_file_box = oasysgui.widgetBox(left_box_3, "", addSpace=False, orientation="vertical")

        file_box = oasysgui.widgetBox(self.b_from_file_box, "", addSpace=True, orientation="horizontal", height=25)

        self.le_file_with_b_vs_y = oasysgui.lineEdit(file_box, self, "file_with_b_vs_y", "File/Url with B vs Y", labelWidth=150, tooltip="file_with_b_vs_y", valueType=str, orientation="horizontal")

        orangegui.button(file_box, self, "...", callback=self.select_file_with_B_vs_Y)

        self.b_from_harmonics_box = oasysgui.widgetBox(left_box_3, "", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.b_from_harmonics_box, self, "id_period", "ID period [m]", labelWidth=260, tooltip="id_period", valueType=float, orientation="horizontal")

        file_box = oasysgui.widgetBox(self.b_from_harmonics_box, "", addSpace=True, orientation="horizontal", height=25)

        self.le_file_with_harmonics = oasysgui.lineEdit(file_box, self, "file_with_harmonics", "File/Url with harmonics", labelWidth=150, tooltip="file_with_harmonics", valueType=str, orientation="horizontal")

        orangegui.button(file_box, self, "...", callback=self.select_file_with_harmonics)

        # Electron initial conditions Box
        left_box_10 = oasysgui.widgetBox(tab_wiggler, "Electron Initial Condition", addSpace=False, orientation="vertical", height=200)

        orangegui.comboBox(left_box_10, self, "shift_betax_flag", tooltip="shift_betax_flag", label="Shift Transversal Velocity", items=["No shift", "Half excursion", "Minimum", "Maximum", "Value at zero", "User value"], callback=self.set_shift_beta_X_flag, labelWidth=260, orientation="horizontal")
        self.shift_betax_value_box = oasysgui.widgetBox(left_box_10, "", addSpace=False, orientation="vertical", height=25)
        self.shift_betax_value_box_hidden = oasysgui.widgetBox(left_box_10, "", addSpace=False, orientation="vertical", height=25)
        oasysgui.lineEdit(self.shift_betax_value_box, self, "shift_betax_value", "Value", tooltip="shift_betax_value", labelWidth=260, valueType=float, orientation="horizontal")

        orangegui.comboBox(left_box_10, self, "shift_x_flag", tooltip="shift_x_flag", label="Shift Transversal Coordinate", items=["No shift", "Half excursion", "Minimum", "Maximum", "Value at zero", "User value"], callback=self.set_shift_X_flag, labelWidth=260, orientation="horizontal")
        self.shift_x_value_box = oasysgui.widgetBox(left_box_10, "", addSpace=False, orientation="vertical", height=25)
        self.shift_x_value_box_hidden = oasysgui.widgetBox(left_box_10, "", addSpace=False, orientation="vertical", height=25)
        oasysgui.lineEdit(self.shift_x_value_box, self, "shift_x_value", "Value [m]", tooltip="shift_x_value", labelWidth=260, valueType=float, orientation="horizontal")

        self.set_shift_X_flag()
        self.set_shift_beta_X_flag()

        # Calculation Box
        left_box_11 = oasysgui.widgetBox(tab_wiggler, "Sampling rays", addSpace=False, orientation="vertical", height=200)

        oasysgui.lineEdit(left_box_11, self, "e_min", "Min photon energy [eV]", tooltip="e_min", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_11, self, "e_max", "Max photon energy [eV]", tooltip="e_max", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_11, self, "number_of_rays", "Number of rays", tooltip="number_of_rays", labelWidth=260, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(left_box_11, self, "seed", "Seed", tooltip="seed", labelWidth=250, valueType=int, orientation="horizontal")

        self.set_shift_X_flag()
        self.set_shift_beta_X_flag()


        # wiggler adv settings
        tab_advanced = oasysgui.createTabPage(self.tabs_control_area, "Advanced Setting")
        left_box_adv = oasysgui.widgetBox(tab_advanced, "Advanced settings", addSpace=False, orientation="vertical", height=200)
        oasysgui.lineEdit(left_box_adv, self, "ng_e", "Number of Points in energy scan", labelWidth=260, tooltip="ng_e", valueType=int, orientation="horizontal")
        oasysgui.lineEdit(left_box_adv, self, "ng_j", "Number of Points in e trajectory (per period)", labelWidth=280, tooltip="ng_j", valueType=int, orientation="horizontal")

        oasysgui.lineEdit(left_box_adv, self, "epsi_dx", "Distance from waist X [m]", labelWidth=260, tooltip="epsi_dx", valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_adv, self, "epsi_dz", "Distance from waist Z [m]", labelWidth=260, tooltip="epsi_dz", valueType=float, orientation="horizontal")


        # wiggler plots
        self.add_specific_wiggler_plots()

        self.set_visibility()

        orangegui.rubber(self.controlArea)

    def add_specific_wiggler_plots(self):
        wiggler_plot_tab = oasysgui.widgetBox(self.main_tabs, addToLayout=0, margin=4)

        self.main_tabs.insertTab(1, wiggler_plot_tab, "Wiggler Plots")

        view_box = oasysgui.widgetBox(wiggler_plot_tab, "Plotting Style", addSpace=False, orientation="horizontal")
        view_box_1 = oasysgui.widgetBox(view_box, "", addSpace=False, orientation="vertical", width=350)

        self.wiggler_view_type_combo = orangegui.comboBox(view_box_1, self,
                                            "plot_wiggler_graph",
                                                          label="Plot Graphs?",
                                                          labelWidth=220,
                                                          items=["No", "Yes"],
                                                          callback=self.refresh_specific_wiggler_plots,
                                                          sendSelectedValue=False,
                                                          orientation="horizontal")

        self.wiggler_tab = []
        self.wiggler_tabs = oasysgui.tabWidget(wiggler_plot_tab)

        current_tab = self.wiggler_tabs.currentIndex()

        size = len(self.wiggler_tab)
        indexes = range(0, size)
        for index in indexes:
            self.wiggler_tabs.removeTab(size-1-index)

        self.wiggler_tab = [
            orangegui.createTabPage(self.wiggler_tabs, "Magnetic Field"),
            orangegui.createTabPage(self.wiggler_tabs, "Electron Curvature"),
            orangegui.createTabPage(self.wiggler_tabs, "Electron Velocity"),
            orangegui.createTabPage(self.wiggler_tabs, "Electron Trajectory"),
            orangegui.createTabPage(self.wiggler_tabs, "Wiggler Spectrum"),
            orangegui.createTabPage(self.wiggler_tabs, "Wiggler Spectral power")
        ]

        for tab in self.wiggler_tab:
            tab.setFixedHeight(self.IMAGE_HEIGHT)
            tab.setFixedWidth(self.IMAGE_WIDTH)

        self.wiggler_plot_canvas = [None, None, None, None, None, None]

        self.wiggler_tabs.setCurrentIndex(current_tab)


    def set_visibility(self):
        self.conventional_sinusoidal_box.setVisible(self.magnetic_field_source == 0)
        self.b_from_file_box.setVisible(self.magnetic_field_source == 1)
        self.b_from_harmonics_box.setVisible(self.magnetic_field_source == 2)

    def select_file_with_B_vs_Y(self):
        self.le_file_with_b_vs_y.setText(oasysgui.selectFileFromDialog(self, self.file_with_b_vs_y, "Open File With B vs Y"))

    def select_file_with_harmonics(self):
        self.le_file_with_harmonics.setText(oasysgui.selectFileFromDialog(self, self.file_with_harmonics, "Open File With Harmonics"))

    def set_shift_X_flag(self):
        self.shift_x_value_box.setVisible(self.shift_x_flag==5)
        self.shift_x_value_box_hidden.setVisible(self.shift_x_flag!=5)

    def set_shift_beta_X_flag(self):
        self.shift_betax_value_box.setVisible(self.shift_betax_flag==5)
        self.shift_betax_value_box_hidden.setVisible(self.shift_betax_flag!=5)

    def get_lightsource(self):
        # syned
        electron_beam = self.get_electron_beam()
        print("\n\n>>>>>> ElectronBeam info: ", electron_beam.info(), type(electron_beam))

        if self.type_of_properties == 3:
            flag_emittance = 0
        else:
            flag_emittance = 1
        # S4Wiggler
        if self.magnetic_field_source == 0:
            sourcewiggler = S4Wiggler(
                    magnetic_field_periodic  = 1,   # 0=external, 1=periodic
                    file_with_magnetic_field = "",  # useful if magnetic_field_periodic=0
                    K_vertical               = self.k_value,
                    period_length            = self.id_period,
                    number_of_periods        = self.number_of_periods, # syned Wiggler pars: useful if magnetic_field_periodic=1
                    emin                     = self.e_min,     # Photon energy scan from energy (in eV)
                    emax                     = self.e_max,     # Photon energy scan to energy (in eV)
                    ng_e                     = self.ng_e,      # Photon energy scan number of points
                    ng_j                     = self.ng_j ,     # Number of points in electron trajectory (per period) for internal calculation only
                    flag_emittance           = flag_emittance, # Use emittance (0=No, 1=Yes)
                    shift_x_flag             = 0,
                    shift_x_value            = 0.0,
                    shift_betax_flag         = 0,
                    shift_betax_value        = 0.0,
                    epsi_dx                  = self.epsi_dx,
                    epsi_dz                  = self.epsi_dz,
                    )

        elif self.magnetic_field_source == 1:
            if self.e_min == self.e_max:
                ng_e = 1
            else:
                ng_e = self.ng_e

            sourcewiggler = S4Wiggler(
                magnetic_field_periodic   = 0,
                file_with_magnetic_field  = self.file_with_b_vs_y,
                emin                      = self.e_min,
                emax                      = self.e_max,
                ng_e                      = ng_e,
                ng_j                      = self.ng_j,
                flag_emittance            = flag_emittance,  # Use emittance (0=No, 1=Yes)
                shift_x_flag              = 4,
                shift_x_value             = 0.0,
                shift_betax_flag          = 4,
                shift_betax_value         = 0.0,
            )
            sourcewiggler.set_electron_initial_conditions_by_label(
                position_label="value_at_zero",
                velocity_label="value_at_zero",
                )

        elif self.magnetic_field_source == 2:
            raise Exception(NotImplemented)

        if self.e_min == self.e_max:
            sourcewiggler.set_energy_monochromatic(self.e_min)

        sourcewiggler.set_electron_initial_conditions(
                        shift_x_flag=self.shift_x_flag,
                        shift_x_value=self.shift_x_value,
                        shift_betax_flag=self.shift_betax_flag,
                        shift_betax_value=self.shift_betax_value)

        print(">>>>>> \n\n S4Wiggler info: ", sourcewiggler.info())

        # S4WigglerLightSource
        lightsource = S4WigglerLightSource(name='wiggler',
                                           electron_beam=electron_beam,
                                           magnetic_structure=sourcewiggler,
                                           nrays=self.number_of_rays,
                                           seed=self.seed)

        print("\n\n>>>>>> S4WigglerLightSource info: ", lightsource.info())

        return lightsource


    def run_shadow4(self):
        sys.stdout = EmittingStream(textWritten=self._write_stdout)

        self._set_plot_quality()

        self.progressBarInit()

        light_source = self.get_lightsource()

        self.progressBarSet(5)
        #
        # run shadow4
        #
        t00 = time.time()
        print(">>>> starting calculation...")
        output_beam = light_source.get_beam()
        photon_energy, flux, spectral_power = light_source.calculate_spectrum()
        t11 = time.time() - t00
        print(">>>> time for %d rays: %f s, %f min, " % (self.number_of_rays, t11, t11 / 60))

        #
        # plots
        #
        self._plot_results(output_beam, None, progressBarValue=80)
        self.refresh_specific_wiggler_plots(light_source, photon_energy, flux, spectral_power)

        #
        # script
        #
        script = light_source.to_python_code()
        script += "\n\n# test plot\nfrom srxraylib.plot.gol import plot_scatter"
        script += "\nrays = beam.get_rays()"
        script += "\nplot_scatter(1e6 * rays[:, 0], 1e6 * rays[:, 2], title='(X,Z) in microns')"

        self.shadow4_script.set_code(script)

        self.progressBarFinished()

        #
        # send beam
        #
        self.send("Shadow Data", ShadowData(beam=output_beam,
                                           number_of_rays=self.number_of_rays,
                                           beamline=S4Beamline(light_source=light_source)))

    def refresh_specific_wiggler_plots(self, lightsource=None, e=None, f=None, w=None):

        if self.plot_wiggler_graph == 0:
            for wiggler_plot_slot_index in range(6):
                current_item = self.wiggler_tab[wiggler_plot_slot_index].layout().itemAt(0)
                self.wiggler_tab[wiggler_plot_slot_index].layout().removeItem(current_item)
                plot_widget_id = oasysgui.QLabel() # TODO: is there a better way to clean this??????????????????????
                self.wiggler_tab[wiggler_plot_slot_index].layout().addWidget(plot_widget_id)
        else:

            if lightsource is None: return
            traj, pars = lightsource.get_trajectory()

            self.plot_widget_item(traj[1, :],traj[7, :],0,
                                  title="Magnetic Field",xtitle="y [m]",ytitle="B [T]")

            self.plot_widget_item(traj[1, :],traj[6, :],1,
                                  title="Electron curvature",xtitle="y [m]",ytitle="cirvature [m^-1]")

            self.plot_widget_item(traj[1, :],traj[3, :],2,
                                  title="Electron velocity",xtitle="y [m]",ytitle="BetaX")

            self.plot_widget_item(traj[1, :],traj[0, :],3,
                                  title="Electron trajectory",xtitle="y [m]",ytitle="x [m]")

            self.plot_widget_item(e,f,4,
                                  title="Wiggler spectrum (current = %5.1f)"%self.ring_current,
                                  xtitle="Photon energy [eV]",ytitle=r"Photons/s/0.1%bw")

            self.plot_widget_item(e,w,5,
                                  title="Wiggler spectrum (current = %5.1f)"%self.ring_current,
                                  xtitle="Photon energy [eV]",ytitle="Spectral power [W/eV]")

    def plot_widget_item(self,x,y,wiggler_plot_slot_index,title="",xtitle="",ytitle=""):

        self.wiggler_tab[wiggler_plot_slot_index].layout().removeItem(self.wiggler_tab[wiggler_plot_slot_index].layout().itemAt(0))
        plot_widget_id = plot_data1D(x.copy(),y.copy(),title=title,xtitle=xtitle,ytitle=ytitle,symbol='.')
        self.wiggler_tab[wiggler_plot_slot_index].layout().addWidget(plot_widget_id)

    def receive_syned_data(self, data):
        sys.stdout = EmittingStream(textWritten=self._write_stdout)
        if data is not None:
            if isinstance(data, Beamline):
                if not data.get_light_source() is None:
                    if isinstance(data.get_light_source().get_magnetic_structure(), InsertionDevice):
                        print(data.get_light_source().get_magnetic_structure(), InsertionDevice)
                        light_source = data.get_light_source()

                        self.magnetic_field_source = 0
                        self.set_visibility()

                        w = light_source.get_magnetic_structure()
                        self.number_of_periods = int(w.number_of_periods())
                        self.id_period = w.period_length()
                        self.k_value = w.K_vertical()

                        self.populate_fields_from_electron_beam(light_source.get_electron_beam())
                    else:
                        raise ValueError("Syned light source not congruent")
                else:
                    raise ValueError("Syned data not correct: light source not present")
            else:
                raise ValueError("Syned data not correct")

if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication
    a = QApplication(sys.argv)
    ow = OWWiggler()
    ow.show()
    a.exec_()





import os, copy, numpy
from shadow4.beam.s4_beam import S4Beam

class ShadowData:
    class ScanningData(object):
        def __init__(self,
                     scanned_variable_name,
                     scanned_variable_value,
                     scanned_variable_display_name,
                     scanned_variable_um,
                     additional_parameters={}):
            self.__scanned_variable_name = scanned_variable_name
            self.__scanned_variable_value = scanned_variable_value
            self.__scanned_variable_display_name = scanned_variable_display_name
            self.__scanned_variable_um = scanned_variable_um
            self.__additional_parameters=additional_parameters

        @property
        def scanned_variable_name(self):
            return self.__scanned_variable_name

        @property
        def scanned_variable_value(self):
            return self.__scanned_variable_value

        @property
        def scanned_variable_display_name(self):
            return self.__scanned_variable_display_name

        @property
        def scanned_variable_um(self):
            return self.__scanned_variable_um

        def has_additional_parameter(self, name):
            return name in self.__additional_parameters.keys()

        def get_additional_parameter(self, name):
            return self.__additional_parameters[name]

    def __init__(self, beam=None, number_of_rays=0, beamline=None):
        if (beam is None):
            if number_of_rays > 0: self.__beam = S4Beam(number_of_rays)
            else:                  self.__beam = S4Beam()
        else:
            self.__beam = beam

        self.__scanning_data = None
        self.__initial_flux  = None
        self.__beamline      = beamline  # added by srio

    @property
    def beam(self):
        return self.__beam

    @beam.setter
    def beam(self, beam):
        self.__beam = beam

    @property
    def beamline(self):
        return self.__beamline

    @beamline.setter
    def beamline(self, beamline):
        self.__beamline = beamline

    @property
    def initial_flux(self):
        return self.__initial_flux

    @initial_flux.setter
    def initial_flux(self, initial_flux):
        self.__initial_flux = initial_flux

    @property
    def scanning_data(self):
        return self.__scanning_data

    @scanning_data.setter
    def scanning_data(self, scanning_data : ScanningData):
        self.__scanning_data = scanning_data

    def get_flux(self, nolost=1):
        if not self.__beam is None and not self.__initial_flux is None:
            return (self.__beam.intensity(nolost) / self.get_number_of_rays(0)) * self.get_initial_flux()
        else:
            return None

    def get_number_of_rays(self, nolost=0):
        if not hasattr(self.__beam, "rays"): return 0
        if nolost == 0:   return self.__beam.rays.shape[0]
        elif nolost == 1: return self.__beam.rays[numpy.where(self.__beam.rays[:, 9] > 0)].shape[0]
        elif nolost == 2: return self.__beam.rays[numpy.where(self.__beam.rays[:, 9] < 0)].shape[0]
        else: raise ValueError("nolost flag value not valid")

    def load_from_file(self, file_name):
        if not self.__beam is None:
            if os.path.exists(file_name): self.__beam.load_h5(file_name)
            else: raise Exception("File " + file_name + " not existing")

    def write_to_file(self, file_name):
        if not self.__beam is None:
            self.__beam.write_h5(file_name)

    def duplicate(self, copy_rays=True):
        beam = S4Beam()
        if copy_rays: beam.rays = copy.deepcopy(self.beam.rays)

        new_shadow_beam = ShadowData(beam=beam)
        new_shadow_beam.scanning_data = self.__scanning_data
        new_shadow_beam.initial_flux  = self.__initial_flux
        new_shadow_beam.beamline = self.__beamline.duplicate()

        return new_shadow_beam

    @classmethod
    def merge_beams(cls, beam_1, beam_2, which_flux=3, merge_history=1):
        if beam_1 and beam_2:
            rays_1 = None
            rays_2 = None

            if len(getattr(beam_1.beam, "rays", numpy.zeros(0))) > 0: rays_1 = copy.deepcopy(beam_1.beam.rays)
            if len(getattr(beam_2.beam, "rays", numpy.zeros(0))) > 0: rays_2 = copy.deepcopy(beam_2.beam.rays)

            merged_beam = beam_1.duplicate(copy_rays=False, history=True)

            merged_beam.oe_number = beam_1.oe_number
            merged_beam.beam.rays = numpy.append(rays_1, rays_2, axis=0)

            merged_beam.beam.rays[:, 11] = numpy.arange(1, len(merged_beam.beam.rays) + 1, 1) # ray_index

            if which_flux ==1 :
                if not beam_1.initial_flux is None:
                    merged_beam.initial_flux = beam_1.initial_flux
            elif which_flux == 2:
                if not beam_2.initial_flux is None:
                    merged_beam.initial_flux = beam_2.initial_flux
            else:
                if not beam_1.initial_flux is None and not beam_2.initial_flux is None:
                    merged_beam.initial_flux = beam_1.initial_flux + beam_2.initial_flux

            if merge_history > 0:
                if beam_1.history and beam_2.history:
                    if len(beam_1.history) == len(beam_2.history):
                        for index in range(1, beam_1.oe_number + 1):
                            history_element_1 =  beam_1.get_oe_history(index)
                            history_element_2 =  beam_2.get_oe_history(index)

                            merged_history_element = merged_beam.get_oe_history(index)
                            merged_history_element.input_data = ShadowData.merge_beams(history_element_1.input_data, history_element_2.input_data, which_flux, merge_history=(merge_history != 1))
                    else:
                        raise ValueError("Histories must have the same path to be merged")
                else:
                    raise ValueError("Both beams must have a history to be merged")

            return merged_beam
        else:
            raise Exception("Both input beams should provided for merging")

    @classmethod
    def initialize_from_beam(cls, input_beam):
        return input_beam.duplicate()




from syned.beamline.shape import Circle
from shadow4.beamline.optical_elements.refractors.s4_lens import S4Lens, S4LensElement
from orangecontrib.shadow4.widgets.gui.ow_abstract_lens import OWAbstractLens

class OWLens(OWAbstractLens):
    name = "Refractive Lens"
    description = "Shadow Refractive Lens"
    icon = "icons/lens.png"
    priority = 2.1

    def __init__(self):
        super().__init__()

    # ----------------------------------------------------
    # from OpticalElement

    def get_optical_element_instance(self):
        try:    name = self.getNode().title
        except: name = "Refractive Lens"

        um_to_si = 1e-6

        if self.has_finite_diameter == 0: boundary_shape = Circle(radius=um_to_si*self.diameter*0.5)
        else:                             boundary_shape = None
        if self.is_cylinder == 1: cylinder_angle = self.cylinder_angle + 1
        else:                     cylinder_angle = 0

        return S4Lens(name=name,
                      boundary_shape=boundary_shape,
                      material="", # not used
                      thickness=self.interthickness*um_to_si,
                      surface_shape=self.surface_shape,
                      convex_to_the_beam=self.convex_to_the_beam,
                      cylinder_angle=cylinder_angle,
                      ri_calculation_mode=self.ri_calculation_mode,
                      prerefl_file=self.prerefl_file,
                      refraction_index=self.refraction_index,
                      attenuation_coefficient=self.attenuation_coefficient,
                      radius=self.radius*um_to_si,
                      conic_coefficients=None) # TODO: add conic coefficient shape to the GUI

    def get_beamline_element_instance(self):
        return S4LensElement()



import sys
import time

from orangecontrib.shadow4.widgets.gui.ow_electron_beam import OWElectronBeam
from orangecontrib.shadow4.widgets.gui.plots import plot_data1D
from orangecontrib.shadow4.util.shadow4_objects import ShadowData

from oasys.widgets import gui as oasysgui

from oasys.util.oasys_util import EmittingStream

from syned.beamline.beamline import Beamline
from syned.storage_ring.magnetic_structures.insertion_device import InsertionDevice
from syned.widget.widget_decorator import WidgetDecorator

from orangewidget import gui as orangegui
from orangewidget.settings import Setting


from shadow4.sources.undulator.s4_undulator import S4Undulator
from shadow4.sources.undulator.s4_undulator_light_source import S4UndulatorLightSource

from shadow4.beamline.s4_beamline import S4Beamline

class OWUndulator(OWElectronBeam, WidgetDecorator):

    name = "Undulator Light Source"
    description = "Undulator Light Source"
    icon = "icons/undulator.png"
    priority = 50

    inputs = []
    WidgetDecorator.append_syned_input_data(inputs)

    outputs = [{"name":"Shadow Data",
                "type":ShadowData,
                "doc":"",}]

    # undulator parameters
    K_vertical = Setting(0.25)  # syned Undulator parameter
    period_length = Setting(0.032)  # syned Undulator parameter
    number_of_periods = Setting(50)  # syned Undulator parameter

    # photon energy
    emin = Setting(10490.0)  # Photon energy scan from energy (in eV)
    emax = Setting(10510.0)  # Photon energy scan to energy (in eV)
    maxangle = Setting(0.015)  # Maximum radiation semiaperture in RADIANS

    # other parameters
    ng_e = Setting(3)  # Photon energy scan number of points
    ng_t = Setting(100)  # Number of points in angle theta
    ng_p = Setting(11)  # Number of points in angle phi
    ng_j = Setting(20)  # Number of points in electron trajectory (per period) for internal calculation only
    code_undul_phot = Setting(0) # "internal",  # internal, pysru, srw
    flag_emittance = Setting(0)  # when sampling rays: Use emittance (0=No, 1=Yes)
    flag_size = Setting(2)  # when sampling rays: 0=point,1=Gaussian,2=FT(Divergences)

    # sampling rays
    number_of_rays = Setting(500)
    seed = Setting(5676561)


    plot_undulator_graph = 1

    beam_out = None


    def __init__(self):
        super().__init__()

        tab_undulator = oasysgui.createTabPage(self.tabs_control_area, "Undulator Setting")

        # undulator parameters box
        left_box_3 = oasysgui.widgetBox(tab_undulator, "Undulator Parameters", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(left_box_3, self, "K_vertical", "K value", labelWidth=260, tooltip="K_vertical", valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_3, self, "period_length", "ID period [m] [m]", labelWidth=260, tooltip="period_length", valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_3, self, "number_of_periods", "Number of Periods", labelWidth=260, tooltip="number_of_periods", valueType=int, orientation="horizontal")



        # photon energy Box
        left_box_10 = oasysgui.widgetBox(tab_undulator, "Sampling intervals", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(left_box_10, self, "emin", "Min photon energy [eV]", labelWidth=260, tooltip="emin", valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_10, self, "emax", "Max photon energy [eV]", labelWidth=260, tooltip="emax", valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_10, self, "maxangle", "Max elevation angle for radiation theta [rad]", labelWidth=260, tooltip="maxangle", valueType=float, orientation="horizontal")


        # internal parameters
        left_box_11 = oasysgui.widgetBox(tab_undulator, "Internal calculation parameters", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(left_box_11, self, "ng_e", "Points in Photon energy (if polychromatic)", tooltip="ng_e", labelWidth=250, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(left_box_11, self, "ng_t", "Points in theta [elevation]", tooltip="ng_t", labelWidth=250, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(left_box_11, self, "ng_p", "Points in phi [azimuthal]", tooltip="ng_p", labelWidth=250, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(left_box_11, self, "ng_j", "Points in electron trajectory", tooltip="ng_j", labelWidth=250, valueType=int, orientation="horizontal")
        orangegui.comboBox(left_box_11, self, "code_undul_phot", label="Use code",
                           items=["internal", "pysru", "srw"], labelWidth=260, orientation="horizontal")
        orangegui.comboBox(left_box_11, self, "flag_size", label="Size sampling in real space",
                           items=["point", "Gaussian", "FT(Divergences)"], labelWidth=260, orientation="horizontal")

        # sampling
        left_box_12 = oasysgui.widgetBox(tab_undulator, "Sampling rays", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(left_box_12, self, "number_of_rays", "Number of rays", labelWidth=260, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(left_box_12, self, "seed", "Seed", tooltip="Seed (0=clock)", labelWidth=250, valueType=int, orientation="horizontal")


        # undulator plots
        self.add_specific_undulator_plots()

        self.set_visibility()

        orangegui.rubber(self.controlArea)


    def add_specific_undulator_plots(self):

        undulator_plot_tab = oasysgui.widgetBox(self.main_tabs, addToLayout=0, margin=4)

        self.main_tabs.insertTab(1, undulator_plot_tab, "Undulator Plots")

        view_box = oasysgui.widgetBox(undulator_plot_tab, "Plotting Style", addSpace=False, orientation="horizontal")
        view_box_1 = oasysgui.widgetBox(view_box, "", addSpace=False, orientation="vertical", width=350)

        self.undulator_view_type_combo = orangegui.comboBox(view_box_1, self,
                                            "plot_undulator_graph",
                                                          label="Plot Graphs?",
                                                          labelWidth=220,
                                                          items=["No", "Yes"],
                                                          callback=self.refresh_specific_undulator_plots,
                                                          sendSelectedValue=False,
                                                          orientation="horizontal")

        self.undulator_tab = []
        self.undulator_tabs = oasysgui.tabWidget(undulator_plot_tab)

        current_tab = self.undulator_tabs.currentIndex()

        size = len(self.undulator_tab)
        indexes = range(0, size)
        for index in indexes:
            self.undulator_tabs.removeTab(size-1-index)

        if self.emin == self.emax:
            self.undulator_tab = [
                orangegui.createTabPage(self.undulator_tabs, "Radiation intensity (polar)"),
                orangegui.createTabPage(self.undulator_tabs, "Polarization (polar)"),
                orangegui.createTabPage(self.undulator_tabs, "Radiation intensity (cartesian - interpolated)"),
                orangegui.createTabPage(self.undulator_tabs, "Photon source size"),
            ]

            self.undulator_plot_canvas = [None,None,None,None,]
        else:
            self.undulator_tab = [
                orangegui.createTabPage(self.undulator_tabs, "Radiation (polar)"),
                orangegui.createTabPage(self.undulator_tabs, "Polarization (polar)"),
                orangegui.createTabPage(self.undulator_tabs, "Radiation (interpolated)"),
                orangegui.createTabPage(self.undulator_tabs, "Photon source size"),
                orangegui.createTabPage(self.undulator_tabs, "Power Density (interpolated)"),
                orangegui.createTabPage(self.undulator_tabs, "Flux"),
                orangegui.createTabPage(self.undulator_tabs, "Spectral Power"),
            ]

            self.undulator_plot_canvas = [None,None,None,None,None,None,None,]

        for tab in self.undulator_tab:
            tab.setFixedHeight(self.IMAGE_HEIGHT)
            tab.setFixedWidth(self.IMAGE_WIDTH)
        # self.undulator_plot_canvas = [None, None, None, None, None, None]

        self.undulator_tabs.setCurrentIndex(current_tab)



    def set_visibility(self):
        pass
        # self.conventional_sinusoidal_box.setVisible(self.magnetic_field_source == 0)
        # self.b_from_file_box.setVisible(self.magnetic_field_source == 1)
        # self.b_from_harmonics_box.setVisible(self.magnetic_field_source == 2)

    # def select_file_with_B_vs_Y(self):
    #     self.le_file_with_b_vs_y.setText(oasysgui.selectFileFromDialog(self, self.file_with_b_vs_y, "Open File With B vs Y"))
    #
    # def select_file_with_harmonics(self):
    #     self.le_file_with_harmonics.setText(oasysgui.selectFileFromDialog(self, self.file_with_harmonics, "Open File With Harmonics"))
    #
    # def set_shift_X_flag(self):
    #     self.shift_x_value_box.setVisible(self.shift_x_flag==5)
    #     self.shift_x_value_box_hidden.setVisible(self.shift_x_flag!=5)
    #
    # def set_shift_beta_X_flag(self):
    #     self.shift_betax_value_box.setVisible(self.shift_betax_flag==5)
    #     self.shift_betax_value_box_hidden.setVisible(self.shift_betax_flag!=5)

    def get_lightsource(self):
        # syned
        electron_beam = self.get_electron_beam()
        print("\n\n>>>>>> ElectronBeam info: ", electron_beam.info(), type(electron_beam))

        ng_j = 501 # trajectory points

        if self.type_of_properties == 3:
            flag_emittance = 0
        else:
            flag_emittance = 1
        # S4undulator

        code_undul_phot = ["internal", "pysru", "srw"][self.code_undul_phot]

        sourceundulator = S4Undulator(
            K_vertical=self.K_vertical,                # syned Undulator parameter
            period_length=self.period_length,          # syned Undulator parameter
            number_of_periods=self.number_of_periods,  # syned Undulator parameter
            emin=self.emin,  # Photon energy scan from energy (in eV)
            emax=self.emax,  # Photon energy scan to energy (in eV)
            ng_e=self.ng_e,  # Photon energy scan number of points
            maxangle=self.maxangle,  # Maximum radiation semiaperture in RADIANS
            ng_t=self.ng_t,  # Number of points in angle theta
            ng_p=self.ng_p,  # Number of points in angle phi
            ng_j=self.ng_j,  # Number of points in electron trajectory (per period) for internal calculation only
            code_undul_phot=code_undul_phot,  # internal, pysru, srw
            flag_emittance=self.flag_emittance,  # when sampling rays: Use emittance (0=No, 1=Yes)
            flag_size=self.flag_size,  # when sampling rays: 0=point,1=Gaussian,2=FT(Divergences)
            )

        print(">>>>>> \n\n S4Undulator info: ", sourceundulator.info())


        # S4undulatorLightSource
        lightsource = S4UndulatorLightSource(name='undulator',
                                           electron_beam=electron_beam,
                                           magnetic_structure=sourceundulator,
                                           nrays=self.number_of_rays,
                                           seed=self.seed)

        print("\n\n>>>>>> S4undulatorLightSource info: ", lightsource.info())

        return lightsource


    def run_shadow4(self):

        sys.stdout = EmittingStream(textWritten=self._write_stdout)

        self._set_plot_quality()

        self.progressBarInit()

        light_source = self.get_lightsource()

        self.progressBarSet(5)
        #
        # run shadow4
        #
        t00 = time.time()
        print(">>>> starting calculation...")
        output_beam = light_source.get_beam()
        # todo:
        # photon_energy, flux, spectral_power = light_source.calculate_spectrum()
        t11 = time.time() - t00
        print(">>>> time for %d rays: %f s, %f min, " % (self.number_of_rays, t11, t11 / 60))


        #
        # plots
        #
        self._plot_results(output_beam, None, progressBarValue=80)
        #todo:
        # self.refresh_specific_undulator_plots(light_source, photon_energy, flux, spectral_power)


        #
        # script
        #
        script = light_source.to_python_code()
        script += "\n\n# test plot\nfrom srxraylib.plot.gol import plot_scatter"
        script += "\nrays = beam.get_rays()"
        script += "\nplot_scatter(1e6 * rays[:, 0], 1e6 * rays[:, 2], title='(X,Z) in microns')"


        self.shadow4_script.set_code(script)

        self.progressBarFinished()

        #
        # send beam
        #
        self.send("Shadow Data", ShadowData(beam=output_beam,
                                           number_of_rays=self.number_of_rays,
                                           beamline=S4Beamline(light_source=light_source)))

    def refresh_specific_undulator_plots(self, lightsource=None, e=None, f=None, w=None):
        pass

        # if self.plot_undulator_graph == 0:
        #     for undulator_plot_slot_index in range(6):
        #         current_item = self.wiggler_tab[wiggler_plot_slot_index].layout().itemAt(0)
        #         self.wiggler_tab[wiggler_plot_slot_index].layout().removeItem(current_item)
        #         plot_widget_id = oasysgui.QLabel() # TODO: is there a better way to clean this??????????????????????
        #         self.wiggler_tab[wiggler_plot_slot_index].layout().addWidget(plot_widget_id)
        # else:
        #
        #     if lightsource is None: return
        #     traj, pars = lightsource.get_trajectory()
        #
        #     self.plot_widget_item(traj[1, :],traj[7, :],0,
        #                           title="Magnetic Field",xtitle="y [m]",ytitle="B [T]")
        #
        #     self.plot_widget_item(traj[1, :],traj[6, :],1,
        #                           title="Electron curvature",xtitle="y [m]",ytitle="cirvature [m^-1]")
        #
        #     self.plot_widget_item(traj[1, :],traj[3, :],2,
        #                           title="Electron velocity",xtitle="y [m]",ytitle="BetaX")
        #
        #     self.plot_widget_item(traj[1, :],traj[0, :],3,
        #                           title="Electron trajectory",xtitle="y [m]",ytitle="x [m]")
        #
        #     self.plot_widget_item(e,f,4,
        #                           title="Wiggler spectrum (current = %5.1f)"%self.ring_current,
        #                           xtitle="Photon energy [eV]",ytitle=r"Photons/s/0.1%bw")
        #
        #     self.plot_widget_item(e,w,5,
        #                           title="Wiggler spectrum (current = %5.1f)"%self.ring_current,
        #                           xtitle="Photon energy [eV]",ytitle="Spectral power [W/eV]")

    def plot_undulator_item(self,x,y,undulator_plot_slot_index,title="",xtitle="",ytitle=""):

        self.undulator_tab[undulator_plot_slot_index].layout().removeItem(self.undulator_tab[undulator_plot_slot_index].layout().itemAt(0))
        plot_widget_id = plot_data1D(x.copy(),y.copy(),title=title,xtitle=xtitle,ytitle=ytitle,symbol='.')
        self.undulator_tab[undulator_plot_slot_index].layout().addWidget(plot_widget_id)

    def receive_syned_data(self, data):
        pass
        # sys.stdout = EmittingStream(textWritten=self.writeStdOut)
        # if data is not None:
        #     if isinstance(data, Beamline):
        #         if not data.get_light_source() is None:
        #             if isinstance(data.get_light_source().get_magnetic_structure(), InsertionDevice):
        #                 print(data.get_light_source().get_magnetic_structure(), InsertionDevice)
        #                 light_source = data.get_light_source()
        #
        #                 self.magnetic_field_source = 0
        #                 self.set_visibility()
        #
        #                 w = light_source.get_magnetic_structure()
        #                 self.number_of_periods = int(w.number_of_periods())
        #                 self.id_period = w.period_length()
        #                 self.k_value = w.K_vertical()
        #
        #                 self.populate_fields_from_electron_beam(light_source.get_electron_beam())
        #
        #             else:
        #                 raise ValueError("Syned light source not congruent")
        #         else:
        #             raise ValueError("Syned data not correct: light source not present")
        #     else:
        #         raise ValueError("Syned data not correct")

if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication
    a = QApplication(sys.argv)
    ow = OWUndulator()
    ow.show()
    a.exec_()




import sys
import time
import numpy

from PyQt5.QtGui import QPalette, QColor, QFont


from orangewidget import gui
from orangewidget import widget
from orangewidget.settings import Setting

from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence
from oasys.util.oasys_util import EmittingStream

from orangecontrib.shadow4.util.shadow4_objects import ShadowData
from orangecontrib.shadow4.widgets.gui.ow_generic_element import GenericElement

from shadow4.beamline.s4_beamline import S4Beamline
from shadow4.sources.source_geometrical.source_geometrical import SourceGeometrical

from syned.widget.widget_decorator import WidgetDecorator


class OWGeometrical(GenericElement, WidgetDecorator):

    name = "Geometrical Source"
    description = "Shadow Source: Geometrical Source"
    icon = "icons/geometrical.png"
    priority = 1

    inputs = []
    WidgetDecorator.append_syned_input_data(inputs)

    outputs = [{"name":"Shadow Data",
                "type":ShadowData,
                "doc":"",}]

    number_of_rays = Setting(5000)
    seed = Setting(5676561)

    spatial_type = Setting(1)

    rect_width = Setting(0.1)
    rect_height = Setting(0.2)
    ell_semiaxis_x = Setting(0.1)
    ell_semiaxis_z = Setting(0.2)
    gauss_sigma_x = Setting(0.001)
    gauss_sigma_z = Setting(0.001)

    angular_distribution = Setting(0)

    horizontal_div_x_plus = Setting(5.0e-7)
    horizontal_div_x_minus = Setting(5.0e-7)
    vertical_div_z_plus = Setting(5.0e-6)
    vertical_div_z_minus = Setting(5.0e-6)

    angular_distribution_limits = Setting(0)

    horizontal_lim_x_plus = Setting(1.0e-5)
    horizontal_lim_x_minus = Setting(1.0e-5)
    vertical_lim_z_plus = Setting(1.0e-5)
    vertical_lim_z_minus = Setting(1.0e-5)
    horizontal_sigma_x = Setting(0.001)
    vertical_sigma_z = Setting(0.0001)

    cone_internal_half_aperture = Setting(0.001)
    cone_external_half_aperture = Setting(0.002)

    depth = Setting(0)

    source_depth_y = Setting(0.002)
    sigma_y = Setting(0.001)

    photon_energy_distribution = Setting(0)

    units=Setting(0)

    single_line_value = Setting(1000.0)
    single_line_value = Setting(1000.0)
    number_of_lines = Setting(0)

    line_value_1 = Setting(1000.0)
    line_value_2 = Setting(1010.0)
    line_value_3 = Setting(0.0)
    line_value_4 = Setting(0.0)
    line_value_5 = Setting(0.0)
    line_value_6 = Setting(0.0)
    line_value_7 = Setting(0.0)
    line_value_8 = Setting(0.0)
    line_value_9 = Setting(0.0)
    line_value_10 = Setting(0.0)

    uniform_minimum = Setting(1000.0)
    uniform_maximum = Setting(1010.0)

    line_int_1 = Setting(0.0)
    line_int_2 = Setting(0.0)
    line_int_3 = Setting(0.0)
    line_int_4 = Setting(0.0)
    line_int_5 = Setting(0.0)
    line_int_6 = Setting(0.0)
    line_int_7 = Setting(0.0)
    line_int_8 = Setting(0.0)
    line_int_9 = Setting(0.0)
    line_int_10 = Setting(0.0)

    gaussian_central_value = Setting(0.0)
    gaussian_sigma = Setting(0.0)
    gaussian_minimum = Setting(0.0)
    gaussian_maximum = Setting(0.0)

    user_defined_file = Setting("energy_spectrum.dat")
    user_defined_minimum = Setting(0.0)
    user_defined_maximum = Setting(0.0)
    user_defined_spectrum_binning = Setting(10000)
    user_defined_refining_factor  = Setting(5)

    # polarization = Setting(1)
    coherent_beam = Setting(0)
    phase_diff = Setting(0.0)
    polarization_degree = Setting(1.0)

    optimize_source=Setting(0)
    optimize_file_name = Setting("NONESPECIFIED")
    max_number_of_rejected_rays = Setting(10000000)


    def __init__(self):
        super().__init__(show_automatic_box=False, has_footprint=False)


        self.runaction = widget.OWAction("Run Shadow4/Source", self)
        self.runaction.triggered.connect(self.run_shadow4)
        self.addAction(self.runaction)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        button = gui.button(button_box, self, "Run shadow4/source", callback=self.run_shadow4)
        font = QFont(button.font())
        font.setBold(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Blue'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)

        button = gui.button(button_box, self, "Reset Fields", callback=self.call_reset_settings)
        font = QFont(button.font())
        font.setItalic(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Red'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)
        button.setFixedWidth(150)

        ################################################################################################################
        self.controlArea.setFixedWidth(self.CONTROL_AREA_WIDTH)

        tabs_setting = oasysgui.tabWidget(self.controlArea)
        tabs_setting.setFixedHeight(self.TABS_AREA_HEIGHT + 60)
        tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH - 5)

        tab_basic = oasysgui.createTabPage(tabs_setting, "Basic Setting")
        tab_geometry = oasysgui.createTabPage(tabs_setting, "Geometry Setting")
        tab_energy = oasysgui.createTabPage(tabs_setting, "Energy/Polarization Setting")

        ##############################
        # MONTECARLO

        left_box_1 = oasysgui.widgetBox(tab_basic, "Montecarlo", addSpace=True, orientation="vertical", height=100)

        gui.separator(left_box_1)

        self.sample_box_1 = oasysgui.widgetBox(left_box_1, "", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.sample_box_1, self, "number_of_rays", "Number of Random Rays", labelWidth=260,
                          valueType=int, orientation="horizontal")
        oasysgui.lineEdit(self.sample_box_1, self, "seed", "Seed (0=clock)", labelWidth=260, valueType=int,
                          orientation="horizontal")

        ##############################
        # GEOMETRY

        left_box_2 = oasysgui.widgetBox(tab_geometry, "", addSpace=True, orientation="vertical", height=550)

        gui.separator(left_box_2)

        ######

        spatial_type_box = oasysgui.widgetBox(left_box_2, "Spatial Type", addSpace=True, orientation="vertical",
                                              height=120)

        items = SourceGeometrical.spatial_type_list() # ["Point", "Rectangle", "Ellipse", "Gaussian"]
        gui.comboBox(spatial_type_box, self, "spatial_type", label="Spatial Type", labelWidth=355,
                     items=items, orientation="horizontal",
                     callback=self.set_SpatialType)

        gui.separator(spatial_type_box)

        self.spatial_type_box_1 = oasysgui.widgetBox(spatial_type_box, "", addSpace=False, orientation="vertical")

        self.le_rect_width = oasysgui.lineEdit(self.spatial_type_box_1, self, "rect_width", "Width X [m]", labelWidth=260,
                                               valueType=float, orientation="horizontal")
        self.le_rect_height = oasysgui.lineEdit(self.spatial_type_box_1, self, "rect_height", "Height Z [m]", labelWidth=260,
                                                valueType=float, orientation="horizontal")

        self.spatial_type_box_2 = oasysgui.widgetBox(spatial_type_box, "", addSpace=False, orientation="vertical")

        self.le_ell_semiaxis_x = oasysgui.lineEdit(self.spatial_type_box_2, self, "ell_semiaxis_x", "Semi-Axis X [m]",
                                                   labelWidth=260, valueType=float, orientation="horizontal")
        self.le_ell_semiaxis_z = oasysgui.lineEdit(self.spatial_type_box_2, self, "ell_semiaxis_z", "Semi-Axis Z [m]",
                                                   labelWidth=260, valueType=float, orientation="horizontal")

        self.spatial_type_box_3 = oasysgui.widgetBox(spatial_type_box, "", addSpace=False, orientation="vertical")

        self.le_gauss_sigma_x = oasysgui.lineEdit(self.spatial_type_box_3, self, "gauss_sigma_x", "Sigma X [m]",
                                                  labelWidth=260, valueType=float, orientation="horizontal")
        self.le_gauss_sigma_z = oasysgui.lineEdit(self.spatial_type_box_3, self, "gauss_sigma_z", "Sigma Z [m]",
                                                  labelWidth=260, valueType=float, orientation="horizontal")

        self.set_SpatialType()

        angular_distribution_box = oasysgui.widgetBox(left_box_2, "Angular Distribution", addSpace=True,
                                                      orientation="vertical", height=260)

        items = SourceGeometrical.angular_distribution_list() # ["Flat", "Uniform", "Gaussian", "Cone","Collimated"]
        gui.comboBox(angular_distribution_box, self, "angular_distribution", label="Angular Distribution",
                     labelWidth=355, items=items, orientation="horizontal", callback=self.set_AngularDistribution)

        gui.separator(angular_distribution_box)

        self.angular_distribution_box_1 = oasysgui.widgetBox(angular_distribution_box, "", addSpace=False,
                                                             orientation="vertical")

        oasysgui.lineEdit(self.angular_distribution_box_1, self, "horizontal_div_x_plus",
                          "Horizontal Divergence X(+) [rad]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.angular_distribution_box_1, self, "horizontal_div_x_minus",
                          "Horizontal Divergence X(-) [rad]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.angular_distribution_box_1, self, "vertical_div_z_plus",
                          "Vertical Divergence Z(+) [rad]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.angular_distribution_box_1, self, "vertical_div_z_minus",
                          "Vertical Divergence Z(-) [rad]", labelWidth=260, valueType=float, orientation="horizontal")

        self.angular_distribution_box_2 = oasysgui.widgetBox(angular_distribution_box, "", addSpace=False,
                                                             orientation="vertical")

        gui.comboBox(self.angular_distribution_box_2, self, "angular_distribution_limits",
                     label="Angular Distribution Limits", labelWidth=355,
                     items=["No Limits", "Horizontal", "Vertical", "Both"], orientation="horizontal",
                     callback=self.set_AngularDistributionLimits)

        self.le_horizontal_lim_x_plus = oasysgui.lineEdit(self.angular_distribution_box_2, self,
                                                          "horizontal_lim_x_plus", "Horizontal Limit X(+) [rad]",
                                                          labelWidth=260, valueType=float, orientation="horizontal")
        self.le_horizontal_lim_x_minus = oasysgui.lineEdit(self.angular_distribution_box_2, self,
                                                           "horizontal_lim_x_minus", "Horizontal Limit X(-) [rad]",
                                                           labelWidth=260, valueType=float, orientation="horizontal")
        self.le_vertical_lim_z_plus = oasysgui.lineEdit(self.angular_distribution_box_2, self, "vertical_lim_z_plus",
                                                        "Vertical Limit Z(+) [rad]", labelWidth=260, valueType=float,
                                                        orientation="horizontal")
        self.le_vertical_lim_z_minus = oasysgui.lineEdit(self.angular_distribution_box_2, self, "vertical_lim_z_minus",
                                                         "Vertical Limit Z(-) [rad]", labelWidth=260, valueType=float,
                                                         orientation="horizontal")

        oasysgui.lineEdit(self.angular_distribution_box_2, self, "horizontal_sigma_x", "Horizontal Sigma (X) [rad]",
                          labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.angular_distribution_box_2, self, "vertical_sigma_z", "Vertical Sigma (Z) [rad]",
                          labelWidth=260, valueType=float, orientation="horizontal")

        self.angular_distribution_box_3 = oasysgui.widgetBox(angular_distribution_box, "", addSpace=False,
                                                             orientation="vertical")

        oasysgui.lineEdit(self.angular_distribution_box_3, self, "cone_internal_half_aperture",
                          "Cone Internal Half-Aperture [rad]", labelWidth=260, valueType=float,
                          orientation="horizontal")
        oasysgui.lineEdit(self.angular_distribution_box_3, self, "cone_external_half_aperture",
                          "Cone External Half-Aperture [rad]", labelWidth=260, valueType=float,
                          orientation="horizontal")

        self.set_AngularDistribution()

        depth_box = oasysgui.widgetBox(left_box_2, "Depth", addSpace=True, orientation="vertical", height=100)

        gui.comboBox(depth_box, self, "depth", label="Depth", labelWidth=355,
                     items=["Off", "Uniform", "Gaussian"], orientation="horizontal", callback=self.set_Depth)

        gui.separator(depth_box, 1)

        self.depth_box_1 = oasysgui.widgetBox(depth_box, "", addSpace=False, orientation="vertical")

        self.le_source_depth_y = oasysgui.lineEdit(self.depth_box_1, self, "source_depth_y", "Source Depth (Y) [m]",
                                                   labelWidth=260, valueType=float, orientation="horizontal")

        self.depth_box_2 = oasysgui.widgetBox(depth_box, "", addSpace=False, orientation="vertical")

        self.le_sigma_y = oasysgui.lineEdit(self.depth_box_2, self, "sigma_y", "Sigma Y [m]", labelWidth=260,
                                            valueType=float, orientation="horizontal")

        self.set_Depth()

        ##############################
        # ENERGY

        left_box_3 = oasysgui.widgetBox(tab_energy, "", addSpace=False, orientation="vertical", height=640)

        ######

        energy_wavelength_box = oasysgui.widgetBox(left_box_3, "Energy/Wavelength", addSpace=False,
                                                   orientation="vertical", height=430)

        items = SourceGeometrical.energy_distribution_list() # ["Single Line", "Several Lines", "Uniform", "Relative Intensities", "Gaussian", "User Defined"]

        gui.comboBox(energy_wavelength_box, self, "photon_energy_distribution", label="Photon Energy Distribution",
                     labelWidth=260, items=items, orientation="horizontal", callback=self.set_PhotonEnergyDistribution)

        gui.comboBox(energy_wavelength_box, self, "units", label="Units", labelWidth=260,
                     items=["Energy/eV", "Wavelength/Å"], orientation="horizontal",
                     callback=self.set_PhotonEnergyDistribution)

        self.ewp_box_5 = oasysgui.widgetBox(energy_wavelength_box, "", addSpace=False, orientation="vertical")

        gui.comboBox(self.ewp_box_5, self, "number_of_lines", label="Number of Lines", labelWidth=330,
                     items=["1", "2", "3", "4", "5", "6", "7", "8", "9", "10"], orientation="horizontal",
                     callback=self.set_NumberOfLines)

        container = oasysgui.widgetBox(energy_wavelength_box, "", addSpace=False, orientation="horizontal")
        self.container_left = oasysgui.widgetBox(container, "", addSpace=False, orientation="vertical")
        self.container_right = oasysgui.widgetBox(container, "", addSpace=False, orientation="vertical")

        self.ewp_box_1 = oasysgui.widgetBox(self.container_left, "", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.ewp_box_1, self, "single_line_value", "Value", labelWidth=260, valueType=float,
                          orientation="horizontal")

        self.ewp_box_2 = oasysgui.widgetBox(self.container_left, "Values", addSpace=False, orientation="vertical")

        self.le_line_value_1 = oasysgui.lineEdit(self.ewp_box_2, self, "line_value_1", "Line 1", valueType=float,
                                                 orientation="horizontal")
        self.le_line_value_2 = oasysgui.lineEdit(self.ewp_box_2, self, "line_value_2", "Line 2", valueType=float,
                                                 orientation="horizontal")
        self.le_line_value_3 = oasysgui.lineEdit(self.ewp_box_2, self, "line_value_3", "Line 3", valueType=float,
                                                 orientation="horizontal")
        self.le_line_value_4 = oasysgui.lineEdit(self.ewp_box_2, self, "line_value_4", "Line 4", valueType=float,
                                                 orientation="horizontal")
        self.le_line_value_5 = oasysgui.lineEdit(self.ewp_box_2, self, "line_value_5", "Line 5", valueType=float,
                                                 orientation="horizontal")
        self.le_line_value_6 = oasysgui.lineEdit(self.ewp_box_2, self, "line_value_6", "Line 6", valueType=float,
                                                 orientation="horizontal")
        self.le_line_value_7 = oasysgui.lineEdit(self.ewp_box_2, self, "line_value_7", "Line 7", valueType=float,
                                                 orientation="horizontal")
        self.le_line_value_8 = oasysgui.lineEdit(self.ewp_box_2, self, "line_value_8", "Line 8", valueType=float,
                                                 orientation="horizontal")
        self.le_line_value_9 = oasysgui.lineEdit(self.ewp_box_2, self, "line_value_9", "Line 9", valueType=float,
                                                 orientation="horizontal")
        self.le_line_value_10 = oasysgui.lineEdit(self.ewp_box_2, self, "line_value_10", "Line 10", valueType=float,
                                                  orientation="horizontal")

        self.ewp_box_3 = oasysgui.widgetBox(self.container_left, "", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.ewp_box_3, self, "uniform_minimum", "Minimum Energy/Wavelength", labelWidth=260,
                          valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.ewp_box_3, self, "uniform_maximum", "Maximum Energy/Wavelength", labelWidth=260,
                          valueType=float, orientation="horizontal")

        self.ewp_box_4 = oasysgui.widgetBox(self.container_right, "Relative Intensities", addSpace=False,
                                            orientation="vertical")

        self.le_line_int_1 = oasysgui.lineEdit(self.ewp_box_4, self, "line_int_1", "Int 1", labelWidth=100,
                                               valueType=float, orientation="horizontal")
        self.le_line_int_2 = oasysgui.lineEdit(self.ewp_box_4, self, "line_int_2", "Int 2", labelWidth=100,
                                               valueType=float, orientation="horizontal")
        self.le_line_int_3 = oasysgui.lineEdit(self.ewp_box_4, self, "line_int_3", "Int 3", labelWidth=100,
                                               valueType=float, orientation="horizontal")
        self.le_line_int_4 = oasysgui.lineEdit(self.ewp_box_4, self, "line_int_4", "Int 4", labelWidth=100,
                                               valueType=float, orientation="horizontal")
        self.le_line_int_5 = oasysgui.lineEdit(self.ewp_box_4, self, "line_int_5", "Int 5", labelWidth=100,
                                               valueType=float, orientation="horizontal")
        self.le_line_int_6 = oasysgui.lineEdit(self.ewp_box_4, self, "line_int_6", "Int 6", labelWidth=100,
                                               valueType=float, orientation="horizontal")
        self.le_line_int_7 = oasysgui.lineEdit(self.ewp_box_4, self, "line_int_7", "Int 7", labelWidth=100,
                                               valueType=float, orientation="horizontal")
        self.le_line_int_8 = oasysgui.lineEdit(self.ewp_box_4, self, "line_int_8", "Int 8", labelWidth=100,
                                               valueType=float, orientation="horizontal")
        self.le_line_int_9 = oasysgui.lineEdit(self.ewp_box_4, self, "line_int_9", "Int 9", labelWidth=100,
                                               valueType=float, orientation="horizontal")
        self.le_line_int_10 = oasysgui.lineEdit(self.ewp_box_4, self, "line_int_10", "Int 10", labelWidth=100,
                                                valueType=float, orientation="horizontal")

        self.ewp_box_6 = oasysgui.widgetBox(energy_wavelength_box, "Gaussian", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.ewp_box_6, self, "gaussian_central_value", "Central Value", labelWidth=260,
                          valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.ewp_box_6, self, "gaussian_sigma", "Sigma", labelWidth=260, valueType=float,
                          orientation="horizontal")

        if False: # not yet implemented... Is that useful?
            gui.separator(self.ewp_box_6)

            oasysgui.lineEdit(self.ewp_box_6, self, "gaussian_minimum", "Minimum Energy/Wavelength", labelWidth=260,
                              valueType=float, orientation="horizontal")
            oasysgui.lineEdit(self.ewp_box_6, self, "gaussian_maximum", "Maximum Energy/Wavelength", labelWidth=260,
                              valueType=float, orientation="horizontal")

        self.ewp_box_7 = oasysgui.widgetBox(energy_wavelength_box, "User Defined", addSpace=False,
                                            orientation="vertical")

        file_box = oasysgui.widgetBox(self.ewp_box_7, "", addSpace=True, orientation="horizontal", height=25)

        self.le_user_defined_file = oasysgui.lineEdit(file_box, self, "user_defined_file", "Spectrum File",
                                                      labelWidth=100, valueType=str, orientation="horizontal")

        gui.button(file_box, self, "...", callback=self.selectFile)

        if False: # why this?
            gui.separator(self.ewp_box_7)

            oasysgui.lineEdit(self.ewp_box_7, self, "user_defined_minimum", "Minimum Energy/Wavelength", labelWidth=260,
                              valueType=float, orientation="horizontal")
            oasysgui.lineEdit(self.ewp_box_7, self, "user_defined_maximum", "Maximum Energy/Wavelength", labelWidth=260,
                              valueType=float, orientation="horizontal")
            oasysgui.lineEdit(self.ewp_box_7, self, "user_defined_spectrum_binning",
                              "Minimum Nr. of Bins of Input Spectrum", labelWidth=260, valueType=int,
                              orientation="horizontal")
            oasysgui.lineEdit(self.ewp_box_7, self, "user_defined_refining_factor", "Refining Factor", labelWidth=260,
                              valueType=int, orientation="horizontal")

        self.set_PhotonEnergyDistribution()

        polarization_box = oasysgui.widgetBox(left_box_3, "Polarization", addSpace=False, orientation="vertical")

        # gui.comboBox(polarization_box, self, "polarization", label="Polarization", labelWidth=310,
        #              items=["No", "Yes"], orientation="horizontal", callback=self.set_Polarization)

        self.ewp_box_8 = oasysgui.widgetBox(polarization_box, "", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(self.ewp_box_8, self, "polarization_degree", "Polarization Degree [cos_s/(cos_s+sin_s)]",
                          labelWidth=310, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.ewp_box_8, self, "phase_diff", "Phase Difference [deg,0=linear,+90=ell/right]",
                          labelWidth=310, valueType=float, orientation="horizontal")
        gui.comboBox(self.ewp_box_8, self, "coherent_beam", label="Coherent Beam", labelWidth=310,
                     items=["No", "Yes"], orientation="horizontal")
        self.ewp_box_8.setVisible(True)

        # self.set_Polarization()

        ##############################

        left_box_4 = oasysgui.widgetBox(tab_basic, "Reject Rays", addSpace=True, orientation="vertical", height=130)

        gui.comboBox(left_box_4, self, "optimize_source", label="Optimize Source",
                     items=["No", "Using file with phase/space volume)", "Using file with slit/acceptance"],
                     labelWidth=120, callback=self.set_OptimizeSource, orientation="horizontal")
        self.optimize_file_name_box = oasysgui.widgetBox(left_box_4, "", addSpace=False, orientation="vertical")

        file_box = oasysgui.widgetBox(self.optimize_file_name_box, "", addSpace=True, orientation="horizontal",
                                      height=25)

        self.le_optimize_file_name = oasysgui.lineEdit(file_box, self, "optimize_file_name", "File Name",
                                                       labelWidth=100, valueType=str, orientation="horizontal")

        gui.button(file_box, self, "...", callback=self.selectOptimizeFile)

        oasysgui.lineEdit(self.optimize_file_name_box, self, "max_number_of_rejected_rays",
                          "Max number of rejected rays (set 0 for infinity)", labelWidth=280, valueType=int,
                          orientation="horizontal")

        self.set_OptimizeSource()

        # adv_other_box = oasysgui.widgetBox(tab_basic, "Optional file output", addSpace=False, orientation="vertical")
        # gui.comboBox(adv_other_box, self, "file_to_write_out", label="Files to write out", labelWidth=120,
        #              items=["None", "Begin.dat", "Debug (begin.dat + start.xx/end.xx)"],
        #              sendSelectedValue=False, orientation="horizontal")

        gui.rubber(self.controlArea)
        gui.rubber(self.mainArea)

    def is_scanning_enabled(self):
        return True

    def call_reset_settings(self):
        super().call_reset_settings()

        self.set_Sampling()
        self.set_SpatialType()
        self.set_AngularDistribution()
        self.set_Depth()
        self.set_PhotonEnergyDistribution()
        # self.set_Polarization()

    def set_OptimizeSource(self):
        self.optimize_file_name_box.setVisible(self.optimize_source != 0)

    def set_SpatialType(self):
        self.spatial_type_box_1.setVisible(self.spatial_type == 1)
        self.spatial_type_box_2.setVisible(self.spatial_type == 2)
        self.spatial_type_box_3.setVisible(self.spatial_type == 3)

    def set_AngularDistributionLimits(self):
        self.le_horizontal_lim_x_plus.setEnabled(self.angular_distribution_limits != 0 and self.angular_distribution_limits != 2)
        self.le_horizontal_lim_x_minus.setEnabled(self.angular_distribution_limits != 0 and self.angular_distribution_limits != 2)
        self.le_vertical_lim_z_plus.setEnabled(self.angular_distribution_limits != 0 and self.angular_distribution_limits != 1)
        self.le_vertical_lim_z_minus.setEnabled(self.angular_distribution_limits != 0 and self.angular_distribution_limits != 1)

    def set_AngularDistribution(self):
        self.angular_distribution_box_1.setVisible(self.angular_distribution == 0 or self.angular_distribution == 1)
        self.angular_distribution_box_2.setVisible(self.angular_distribution == 2)
        self.angular_distribution_box_3.setVisible(self.angular_distribution == 3)

        if self.angular_distribution == 2:
            self.set_AngularDistributionLimits()

    def set_Depth(self):
        self.depth_box_1.setVisible(self.depth == 1)
        self.depth_box_2.setVisible(self.depth == 2)

    def set_PhotonEnergyDistribution(self):
        self.ewp_box_1.setVisible(self.photon_energy_distribution == 0)
        self.ewp_box_2.setVisible(self.photon_energy_distribution == 1 or self.photon_energy_distribution == 3)
        self.ewp_box_3.setVisible(self.photon_energy_distribution == 2)
        self.ewp_box_4.setVisible(self.photon_energy_distribution == 3)
        self.ewp_box_5.setVisible(self.photon_energy_distribution == 1 or self.photon_energy_distribution == 3)
        self.ewp_box_6.setVisible(self.photon_energy_distribution == 4)
        self.ewp_box_7.setVisible(self.photon_energy_distribution == 5)

        if self.photon_energy_distribution == 3:
            self.le_line_value_1.parentWidget().children()[1].setFixedWidth(100)
            self.le_line_value_2.parentWidget().children()[1].setFixedWidth(100)
            self.le_line_value_3.parentWidget().children()[1].setFixedWidth(100)
            self.le_line_value_4.parentWidget().children()[1].setFixedWidth(100)
            self.le_line_value_5.parentWidget().children()[1].setFixedWidth(100)
            self.le_line_value_6.parentWidget().children()[1].setFixedWidth(100)
            self.le_line_value_7.parentWidget().children()[1].setFixedWidth(100)
            self.le_line_value_8.parentWidget().children()[1].setFixedWidth(100)
            self.le_line_value_9.parentWidget().children()[1].setFixedWidth(100)
            self.le_line_value_10.parentWidget().children()[1].setFixedWidth(100)
        else:
            self.le_line_value_1.parentWidget().children()[1].setFixedWidth(260)
            self.le_line_value_2.parentWidget().children()[1].setFixedWidth(260)
            self.le_line_value_3.parentWidget().children()[1].setFixedWidth(260)
            self.le_line_value_4.parentWidget().children()[1].setFixedWidth(260)
            self.le_line_value_5.parentWidget().children()[1].setFixedWidth(260)
            self.le_line_value_6.parentWidget().children()[1].setFixedWidth(260)
            self.le_line_value_7.parentWidget().children()[1].setFixedWidth(260)
            self.le_line_value_8.parentWidget().children()[1].setFixedWidth(260)
            self.le_line_value_9.parentWidget().children()[1].setFixedWidth(260)
            self.le_line_value_10.parentWidget().children()[1].setFixedWidth(260)

        self.container_right.setVisible(self.photon_energy_distribution == 3)

        self.set_NumberOfLines()

    def set_NumberOfLines(self):
        self.le_line_value_2.parentWidget().setVisible(self.number_of_lines >= 1)
        self.le_line_int_2.parentWidget().setVisible(self.number_of_lines >= 1)
        self.le_line_value_3.parentWidget().setVisible(self.number_of_lines >= 2)
        self.le_line_int_3.parentWidget().setVisible(self.number_of_lines >= 2)
        self.le_line_value_4.parentWidget().setVisible(self.number_of_lines >= 3)
        self.le_line_int_4.parentWidget().setVisible(self.number_of_lines >= 3)
        self.le_line_value_5.parentWidget().setVisible(self.number_of_lines >= 4)
        self.le_line_int_5.parentWidget().setVisible(self.number_of_lines >= 4)
        self.le_line_value_6.parentWidget().setVisible(self.number_of_lines >= 5)
        self.le_line_int_6.parentWidget().setVisible(self.number_of_lines >= 5)
        self.le_line_value_7.parentWidget().setVisible(self.number_of_lines >= 6)
        self.le_line_int_7.parentWidget().setVisible(self.number_of_lines >= 6)
        self.le_line_value_8.parentWidget().setVisible(self.number_of_lines >= 7)
        self.le_line_int_8.parentWidget().setVisible(self.number_of_lines >= 7)
        self.le_line_value_9.parentWidget().setVisible(self.number_of_lines >= 8)
        self.le_line_int_9.parentWidget().setVisible(self.number_of_lines >= 8)
        self.le_line_value_10.parentWidget().setVisible(self.number_of_lines == 9)
        self.le_line_int_10.parentWidget().setVisible(self.number_of_lines == 9)

    # def set_Polarization(self):
    #     self.ewp_box_8.setVisible(self.polarization==1)

    def selectFile(self):
        self.le_user_defined_file.setText(oasysgui.selectFileFromDialog(self, self.user_defined_file, "Open Spectrum File", file_extension_filter="Data Files (*.dat *.txt)"))

    def selectOptimizeFile(self):
        self.le_optimize_file_name.setText(oasysgui.selectFileFromDialog(self, self.optimize_file_name, "Open Optimize Source Parameters File"))


    def checkFields(self):
        # TODO: complete?
        self.number_of_rays = congruence.checkPositiveNumber(self.number_of_rays, "Number of rays")
        self.seed = congruence.checkPositiveNumber(self.seed, "Seed")
        self.energy = congruence.checkPositiveNumber(self.energy, "Energy")
        self.delta_e = congruence.checkPositiveNumber(self.delta_e, "Delta Energy")
        self.undulator_length = congruence.checkPositiveNumber(self.undulator_length, "Undulator Length")


    def get_lightsource(self):

        gs = SourceGeometrical(name="SourceGeometrical", nrays=self.number_of_rays, seed=self.seed)

        if self.spatial_type == 0: # point
            gs.set_spatial_type_point()
        elif self.spatial_type == 1: # rectangle
            gs.set_spatial_type_rectangle(width=self.rect_width,
                                          height=self.rect_height,
                                          )
        elif self.spatial_type == 2: # ellipse
            gs.set_spatial_type_ellipse(width=2*self.ell_semiaxis_x,
                                        height=2*self.ell_semiaxis_z,
                                        )
        elif self.spatial_type == 3: # Gaussian
            gs.set_spatial_type_gaussian(sigma_h=self.gauss_sigma_x,
                                         sigma_v=self.gauss_sigma_z,
                                         )

        if self.depth == 0:
            gs.set_depth_distribution_off()
        elif self.depth == 1:
            gs.set_depth_distribution_uniform(self.source_depth_y)
        elif self.depth == 2:
            gs.set_depth_distribution_gaussian(self.sigma_y)


        if self.angular_distribution == 0: # flat
            gs.set_angular_distribution_flat(hdiv1=-self.horizontal_div_x_minus,
                                             hdiv2=self.horizontal_div_x_plus,
                                             vdiv1=-self.vertical_div_z_minus,
                                             vdiv2=self.vertical_div_z_plus,
                                               )
        elif self.angular_distribution == 1: # Uniform
            gs.set_angular_distribution_uniform(hdiv1=-self.horizontal_div_x_minus,
                                                hdiv2=self.horizontal_div_x_plus,
                                                vdiv1=-self.vertical_div_z_minus,
                                                vdiv2=self.vertical_div_z_plus,
                                               )
        elif self.angular_distribution == 2:  # Gaussian
            if self.angular_distribution_limits != 0:
                raise Exception("Not yet implemented")
            gs.set_angular_distribution_gaussian(sigdix=self.horizontal_sigma_x,
                                                 sigdiz=self.vertical_sigma_z,
                                                 )
        elif self.angular_distribution == 3:  # cone
            gs.set_angular_distribution_cone(cone_max=self.cone_external_half_aperture,
                                                cone_min=self.cone_internal_half_aperture,
                                                )

        elif self.angular_distribution == 4:  # Zero (collimated) - New in shadow4
            gs.set_angular_distribution_collimated()


        # photon energy
        values = [self.line_value_1,
                  self.line_value_2,
                  self.line_value_3,
                  self.line_value_4,
                  self.line_value_5,
                  self.line_value_6,
                  self.line_value_7,
                  self.line_value_8,
                  self.line_value_9,
                  self.line_value_10,
                  ]

        weights = [self.line_int_1,
                  self.line_int_2,
                  self.line_int_3,
                  self.line_int_4,
                  self.line_int_5,
                  self.line_int_6,
                  self.line_int_7,
                  self.line_int_8,
                  self.line_int_9,
                  self.line_int_10,
                  ]

        values = values[0:(self.number_of_lines+1)]
        weights = weights[0:(self.number_of_lines+1)]
        unit = ['eV','A'][self.units]

        if self.photon_energy_distribution == 0: # "Single line":
            gs.set_energy_distribution_singleline(self.single_line_value, unit=unit)
        elif self.photon_energy_distribution == 1: #"Several lines":
            gs.set_energy_distribution_severallines(values=values, unit=unit)
        elif self.photon_energy_distribution == 2: # "Uniform":
            gs.set_energy_distribution_uniform(value_min=self.uniform_minimum,value_max=self.uniform_maximum,unit=unit)
        elif self.photon_energy_distribution == 3: # "Relative intensities":
            gs.set_energy_distribution_relativeintensities(values=values,weights=weights,unit=unit)
        elif self.photon_energy_distribution == 4: # "Gaussian":
            gs.set_energy_distribution_gaussian(center=self.gaussian_central_value,sigma=self.gaussian_sigma,unit=unit)
        elif self.photon_energy_distribution == 5: # "User defined":
            a = numpy.loadtxt(self.user_defined_file)
            gs.set_energy_distribution_userdefined(a[:,0],a[:,1],unit=unit)


        # polarization / coherence
        gs.set_polarization(polarization_degree=self.polarization_degree,
                            phase_diff=self.phase_diff,
                            coherent_beam=self.coherent_beam)

        return gs

    def run_shadow4(self):

        sys.stdout = EmittingStream(textWritten=self._write_stdout)

        self._set_plot_quality()

        self.progressBarInit()

        light_source = self.get_lightsource()

        self.progressBarSet(5)

        # run shadow4

        t00 = time.time()
        # beam = light_source.get_beam(NRAYS=self.number_of_rays, SEED=self.seed)
        output_beam = light_source.get_beam()
        t11 = time.time() - t00
        print(">>>> time for %d rays: %f s, %f min, " % (self.number_of_rays, t11, t11 / 60))

        #
        # beam plots
        #
        self._plot_results(output_beam, None, progressBarValue=80)

        #
        # script
        #
        script = light_source.to_python_code()
        # script += "\n\n\n# run shadow4"
        # script += "\nbeam = light_source.get_beam(N=%d, ISTAR1=%d)" % \
        #           (self.number_of_rays, self.seed)

        script += "\n\n# test plot\nfrom srxraylib.plot.gol import plot_scatter"
        script += "\nrays = beam.get_rays()"
        script += "\nplot_scatter(1e6 * rays[:, 0], 1e6 * rays[:, 2], title='(X,Z) in microns')"

        self.shadow4_script.set_code(script)

        self.progressBarFinished()

        #
        # send beam
        #
        self.send("Shadow Data", ShadowData(beam=output_beam,
                                           number_of_rays=self.number_of_rays,
                                           beamline=S4Beamline(light_source=light_source)))
if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication
    a = QApplication(sys.argv)
    ow = OWGeometrical()
    ow.show()
    a.exec_()
    ow.saveSettings()



import numpy

from orangewidget import gui
from orangewidget.settings import Setting

from oasys.widgets import gui as oasysgui

from syned.beamline.element_coordinates import ElementCoordinates
from syned.beamline.shape import Rectangle
from syned.beamline.shape import Ellipse

from shadow4.beamline.optical_elements.absorbers.s4_screen import S4ScreenElement, S4Screen

from orangecontrib.shadow4.widgets.gui.ow_optical_element import OWOpticalElement, NO_FILE_SPECIFIED

from orangecanvas.resources import icon_loader
from orangecanvas.scheme.node import SchemeNode


class OWScreenSlits(OWOpticalElement):
    name        = "Generic Beam Screen/Slit/Stopper/Attenuator"
    description = "Shadow Screen/Slit/Stopper/Attenuator"
    icon        = "icons/generic_beam_stopper.png"

    priority = 1.1

    aperturing           = Setting(0)
    open_slit_solid_stop = Setting(0)
    aperture_shape       = Setting(0)
    absorption           = Setting(0)
    slit_width_xaxis     = Setting(0.0)
    slit_height_zaxis    = Setting(0.0)
    slit_center_xaxis    = Setting(0.0)
    slit_center_zaxis    = Setting(0.0)
    thickness            = Setting(0.0)
    opt_const_file_name  = Setting(NO_FILE_SPECIFIED)
    material             = Setting("Au")
    density              = Setting(19.3)

    def createdFromNode(self, node):
        super(OWScreenSlits, self).createdFromNode(node)
        self.__change_icon_from_oe_type()

    def widgetNodeAdded(self, node_item : SchemeNode):
        super(OWScreenSlits, self).widgetNodeAdded(node_item)
        self.__change_icon_from_oe_type()

    def __change_icon_from_oe_type(self):
        try:
            title, icon = self.title_and_icon_for_oe

            node = self.getNode()
            node.description.icon = icon
            self.changeNodeIcon(icon_loader.from_description(node.description).get(node.description.icon))
            if node.title in self.oe_names: self.changeNodeTitle(title)
        except:
            pass

    @property
    def oe_names(self):
        return ["Generic Beam Stopper", "Screen", "Aperture", "Obstruction", "Absorber", "Aperture/Absorber", "Obstruction/Absorber"]

    @property
    def title_and_icon_for_oe(self):
        if self.absorption == 0:
            if self.aperturing == 0: return self.oe_names[1], "icons/screen.png"
            elif self.aperturing == 1:
                if self.open_slit_solid_stop == 0: return self.oe_names[2], "icons/aperture_only.png"
                else:                              return self.oe_names[3], "icons/obstruction_only.png"
        else:
            if self.aperturing == 0: return self.oe_names[4], "icons/absorber.png"
            elif self.aperturing == 1:
                if self.open_slit_solid_stop == 0: return self.oe_names[5], "icons/aperture_absorber.png"
                else:                              return self.oe_names[6], "icons/obstruction_absorber.png"

    def __init__(self):
        super().__init__(has_footprint=False)

    def populate_tab_position(self, tab_position):
        self.orientation_box = oasysgui.widgetBox(tab_position, "Optical Element Orientation", addSpace=True, orientation="vertical")

        oasysgui.lineEdit(self.orientation_box, self, "source_plane_distance", "Source Plane Distance [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="source_plane_distance")
        oasysgui.lineEdit(self.orientation_box, self, "image_plane_distance", "Image Plane Distance [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="image_plane_distance")

    def create_basic_settings_subtabs(self, tabs_basic_settings):
        return oasysgui.createTabPage(tabs_basic_settings, "Beam Stopper Type")  # to be populated

    def populate_basic_setting_subtabs(self, basic_setting_subtabs):
        tab_beam_stopper_type = basic_setting_subtabs

        box_aperturing = oasysgui.widgetBox(tab_beam_stopper_type, "Screen/Slit Shape", addSpace=False, orientation="vertical", height=240)

        gui.comboBox(box_aperturing, self, "aperturing", label="Aperturing", labelWidth=350,
                     items=["No", "Yes"],
                     callback=self.set_aperturing, sendSelectedValue=False, orientation="horizontal")

        gui.separator(box_aperturing)

        self.box_aperturing_shape = oasysgui.widgetBox(box_aperturing, "", addSpace=False, orientation="vertical")

        gui.comboBox(self.box_aperturing_shape, self, "open_slit_solid_stop", label="Open slit/Solid stop", labelWidth=260,
                     items=["aperture/slit", "obstruction/stop"],
                     callback=self.set_open_slit_solid_stop, sendSelectedValue=False, orientation="horizontal")

        gui.comboBox(self.box_aperturing_shape, self, "aperture_shape", label="Aperture shape", labelWidth=260,
                     items=["Rectangular", "Ellipse"],
                     sendSelectedValue=False, orientation="horizontal")

        box_aperturing_shape = oasysgui.widgetBox(self.box_aperturing_shape, "", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(box_aperturing_shape, self, "slit_width_xaxis", "Slit width/x-axis   [m]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_aperturing_shape, self, "slit_height_zaxis", "Slit height/z-axis [m]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_aperturing_shape, self, "slit_center_xaxis", "Slit center/x-axis [m]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_aperturing_shape, self, "slit_center_zaxis", "Slit center/z-axis [m]", labelWidth=260, valueType=float, orientation="horizontal")

        box_absorption = oasysgui.widgetBox(tab_beam_stopper_type, "Absorption Parameters", addSpace=False, orientation="vertical", height=200)

        gui.comboBox(box_absorption, self, "absorption", label="Absorption", labelWidth=350,
                     items=["No", "Yes (using preprocessor file)", "Yes (using xraylib)", "Yes (using dabax)"],
                     callback=self.set_absorption, sendSelectedValue=False, orientation="horizontal")

        gui.separator(box_absorption)#, width=self.INNER_BOX_WIDTH_L0)
        self.box_thickness = oasysgui.widgetBox(box_absorption, "", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.box_thickness, self, "thickness", "Thickness [m]", labelWidth=180, valueType=float, orientation="horizontal")

        self.box_absorption       = oasysgui.widgetBox(box_absorption, "", addSpace=False, orientation="vertical")

        file_box = oasysgui.widgetBox(self.box_absorption, "", addSpace=False, orientation="horizontal", height=25)

        self.le_opt_const_file_name = oasysgui.lineEdit(file_box, self, "opt_const_file_name", "prerefl file", labelWidth=130, valueType=str, orientation="horizontal")

        gui.button(file_box, self, "...", callback=self.select_opt_const_file_name)

        self.box_material = oasysgui.widgetBox(box_absorption, "", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(self.box_material, self, "material",
                          "material (formula): ", labelWidth=180, valueType=str,
                          orientation="horizontal", tooltip="material")


        oasysgui.lineEdit(self.box_material, self, "density",
                          "density [g/cm^3]: ", labelWidth=180, valueType=float,
                          orientation="horizontal", tooltip="density")

        self.set_aperturing(is_init=True)
        self.set_absorption(is_init=True)

    def set_aperturing(self, is_init=False):
        self.box_aperturing_shape.setVisible(self.aperturing == 1)

        if not is_init: self.__change_icon_from_oe_type()

    def set_open_slit_solid_stop(self, is_init=False):
        if not is_init: self.__change_icon_from_oe_type()

    def set_absorption(self, is_init=False):
        #self.box_absorption_empty.setVisible(self.absorption == 0)
        self.box_thickness.setVisible(self.absorption >= 1)
        self.box_absorption.setVisible(self.absorption == 1)
        self.box_material.setVisible(self.absorption in (2,3))

        if not is_init: self.__change_icon_from_oe_type()

    def select_opt_const_file_name(self):
        self.le_opt_const_file_name.setText(oasysgui.selectFileFromDialog(self, self.opt_const_file_name, "Open Opt. Const. File"))

    # ----------------------------------------------------
    # from OpticalElement

    def get_coordinates_instance(self):
        return ElementCoordinates(
                p=self.source_plane_distance,
                q=self.image_plane_distance,
                angle_radial=0.0,
                angle_azimuthal=0.0,
                angle_radial_out=numpy.pi,
                )

    def get_optical_element_instance(self):
        boundary_shape = None

        if self.aperturing == 1:
            if self.aperture_shape == 0:   boundary_shape = Rectangle(x_left=self.slit_center_xaxis - self.slit_width_xaxis*0.5,
                                                                      x_right=self.slit_center_xaxis + self.slit_width_xaxis*0.5,
                                                                      y_bottom=self.slit_center_zaxis - self.slit_height_zaxis*0.5,
                                                                      y_top=self.slit_center_zaxis + self.slit_height_zaxis*0.5)
            elif self.aperture_shape == 1: boundary_shape = Ellipse(a_axis_min=self.slit_center_xaxis - self.slit_width_xaxis*0.5,
                                                                    a_axis_max=self.slit_center_xaxis + self.slit_width_xaxis*0.5,
                                                                    b_axis_min=self.slit_center_zaxis - self.slit_height_zaxis*0.5,
                                                                    b_axis_max=self.slit_center_zaxis + self.slit_height_zaxis*0.5)

        return S4Screen(name=self.getNode().title,
                        boundary_shape=boundary_shape,
                        i_abs=self.absorption,
                        i_stop=self.open_slit_solid_stop==1,
                        thick=self.thickness,
                        file_abs=self.opt_const_file_name,
                        material=self.material,
                        density=self.density)

    def get_beamline_element_instance(self): return S4ScreenElement()

if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication
    import sys
    a = QApplication(sys.argv)
    ow = OWScreenSlits()
    ow.show()
    a.exec_()
    ow.saveSettings()



import sys
from PyQt5.QtGui import QPalette, QColor, QFont

from orangewidget import gui
from orangewidget import widget
from orangewidget.settings import Setting

from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence
from oasys.util.oasys_util import EmittingStream

from orangecontrib.shadow4.util.shadow4_objects import ShadowData
from orangecontrib.shadow4.widgets.gui.ow_generic_element import GenericElement

from syned.widget.widget_decorator import WidgetDecorator

from shadow4.beamline.s4_beamline import S4Beamline
from shadow4.sources.source_geometrical.source_grid_polar import SourceGridPolar
from shadow4.sources.source_geometrical.source_grid_cartesian import SourceGridCartesian


class OWGrid(GenericElement, WidgetDecorator):

    name = "Grid Source"
    description = "Shadow Source: Grid Source"
    icon = "icons/grid.png"
    priority = 2

    inputs = []
    WidgetDecorator.append_syned_input_data(inputs)

    outputs = [{"name":"Shadow Data",
                "type":ShadowData,
                "doc":"",}]


    # for both grid and cartesian
    coordinates = Setting(1)

    real_space_width_x = Setting(2e-3)
    real_space_width_z = Setting(2e-3)
    real_space_center_x = Setting(0.0)
    real_space_center_z = Setting(0.0)

    direction_space_width_x = Setting(20e-3)
    direction_space_width_z = Setting(20e-3)
    direction_space_center_x = Setting(0.0)
    direction_space_center_z = Setting(0.0)

    # for grid
    real_space_points_r = Setting(2)
    real_space_points_theta = Setting(8)

    direction_space_points_r = Setting(3)
    direction_space_points_theta = Setting(359)

    # for cartesian
    direction_space_width_y = Setting(20e-3)
    real_space_points_x = Setting(10)
    real_space_points_y = Setting(10)
    real_space_points_z = Setting(10)

    direction_space_points_x = Setting(1)
    direction_space_points_z = Setting(1)


    units=Setting(0)
    single_line_value = Setting(1000.0)

    # polarization = Setting(1)
    polarization_phase_deg = Setting(0.0)
    polarization_degree = Setting(1.0)


    def __init__(self):
        super().__init__(show_automatic_box=False, has_footprint=False)


        self.runaction = widget.OWAction("Run Shadow4/Source", self)
        self.runaction.triggered.connect(self.run_shadow4)
        self.addAction(self.runaction)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        button = gui.button(button_box, self, "Run shadow4/source", callback=self.run_shadow4)
        font = QFont(button.font())
        font.setBold(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Blue'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)

        button = gui.button(button_box, self, "Reset Fields", callback=self.call_reset_settings)
        font = QFont(button.font())
        font.setItalic(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Red'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)
        button.setFixedWidth(150)

        ################################################################################################################
        self.controlArea.setFixedWidth(self.CONTROL_AREA_WIDTH)

        tabs_setting = oasysgui.tabWidget(self.controlArea)
        tabs_setting.setFixedHeight(self.TABS_AREA_HEIGHT + 60)
        tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH - 5)

        tab_basic = oasysgui.createTabPage(tabs_setting, "Basic Setting")
        tab_geometry = oasysgui.createTabPage(tabs_setting, "Geometry Setting")
        tab_energy = oasysgui.createTabPage(tabs_setting, "Energy/Polarization Setting")

        ##############################
        # BASIC

        left_box_1 = oasysgui.widgetBox(tab_basic, "Coordinates", addSpace=True, orientation="vertical")
        gui.comboBox(left_box_1, self, "coordinates", label="Coordinates", labelWidth=355,
                     items=["Cartesian", "Polar"], orientation="horizontal",
                     callback=self.set_coordinates_visibility)

        #### points
        points_box = oasysgui.widgetBox(tab_basic, "Number of points", addSpace=True,
                                                   orientation="vertical")
        ##
        self.points_polar_box = oasysgui.widgetBox(points_box, "", addSpace=True, orientation="vertical")

        oasysgui.lineEdit(self.points_polar_box, self, "real_space_points_r",
                          "Real space Radial points", tooltip="real_space_points_r", labelWidth=260, valueType=int,
                          orientation="horizontal")

        oasysgui.lineEdit(self.points_polar_box, self, "real_space_points_theta",
                          "Real Space Azimuthal points", tooltip="real_space_points_theta", labelWidth=260, valueType=int,
                          orientation="horizontal")

        oasysgui.lineEdit(self.points_polar_box, self, "direction_space_points_r",
                          "Direction space Radial points", tooltip="direction_space_points_r", labelWidth=260, valueType=int,
                          orientation="horizontal")

        oasysgui.lineEdit(self.points_polar_box, self, "direction_space_points_theta",
                          "Direction space Azimuthal points", tooltip="direction_space_points_theta", labelWidth=260, valueType=int,
                          orientation="horizontal")

        ##
        self.points_cartesian_box = oasysgui.widgetBox(points_box, "", addSpace=True, orientation="vertical")

        oasysgui.lineEdit(self.points_cartesian_box, self, "real_space_points_x",
                          "Real space X points", tooltip="real_space_points_x", labelWidth=260, valueType=int,
                          orientation="horizontal")

        oasysgui.lineEdit(self.points_cartesian_box, self, "real_space_points_z",
                          "Real Space Z points", tooltip="real_space_points_z", labelWidth=260, valueType=int,
                          orientation="horizontal")

        oasysgui.lineEdit(self.points_cartesian_box, self, "direction_space_points_x",
                          "Direction space X' points", tooltip="direction_space_points_x", labelWidth=260, valueType=int,
                          orientation="horizontal")

        oasysgui.lineEdit(self.points_cartesian_box, self, "direction_space_points_z",
                          "Direction space Z' points", tooltip="direction_space_points_z", labelWidth=260, valueType=int,
                          orientation="horizontal")
        ##############################
        # GEOMETRY

        left_box_2 = oasysgui.widgetBox(tab_geometry, "", addSpace=True, orientation="vertical", height=550)

        ###### real space
        real_distribution_box = oasysgui.widgetBox(left_box_2, "Real space", addSpace=True,
                                                      orientation="vertical") #, height=260)

        oasysgui.lineEdit(real_distribution_box, self, "real_space_center_x",
                          "Center in X [m]", tooltip="real_space_center_x", labelWidth=260, valueType=float,
                          orientation="horizontal")

        oasysgui.lineEdit(real_distribution_box, self, "real_space_center_z",
                          "Center in Z [m]", tooltip="real_space_center_z", labelWidth=260, valueType=float,
                          orientation="horizontal")

        oasysgui.lineEdit(real_distribution_box, self, "real_space_width_x",
                          "Width in X [m]", tooltip="real_space_width_x", labelWidth=260, valueType=float, orientation="horizontal")

        oasysgui.lineEdit(real_distribution_box, self, "real_space_width_z",
                          "Width in Z [m]", tooltip="real_space_width_z", labelWidth=260, valueType=float, orientation="horizontal")

        gui.separator(left_box_2)

        ###### direction space
        angular_distribution_box = oasysgui.widgetBox(left_box_2, "Direction space (divergences)", addSpace=True,
                                                      orientation="vertical") #, height=260)

        oasysgui.lineEdit(angular_distribution_box, self, "direction_space_center_x",
                          "Center in X' [rad]", tooltip="direction_space_center_x", labelWidth=260, valueType=float,
                          orientation="horizontal")

        oasysgui.lineEdit(angular_distribution_box, self, "direction_space_center_z",
                          "Center in Z' [rad]", tooltip="direction_space_center_z", labelWidth=260, valueType=float,
                          orientation="horizontal")

        oasysgui.lineEdit(angular_distribution_box, self, "direction_space_width_x",
                          "Width in X' [rad]", tooltip="direction_space_width_x", labelWidth=260, valueType=float, orientation="horizontal")

        oasysgui.lineEdit(angular_distribution_box, self, "direction_space_width_z",
                          "Width in Z' [rad]", tooltip="direction_space_width_z", labelWidth=260, valueType=float, orientation="horizontal")



        gui.separator(left_box_2)


        ##############################
        # ENERGY

        left_box_3 = oasysgui.widgetBox(tab_energy, "", addSpace=False, orientation="vertical") #, height=640)

        energy_wavelength_box = oasysgui.widgetBox(left_box_3, "Energy/Wavelength", addSpace=False,
                                                   orientation="vertical")

        gui.comboBox(energy_wavelength_box, self, "units", label="Units", labelWidth=260,
                     items=["Energy/eV", "Wavelength/Å"], orientation="horizontal")

        oasysgui.lineEdit(energy_wavelength_box, self, "single_line_value", "Value", tooltip="single_line_value", labelWidth=260, valueType=float,
                          orientation="horizontal")


        polarization_box = oasysgui.widgetBox(left_box_3, "Polarization", addSpace=False, orientation="vertical")

        self.ewp_box_8 = oasysgui.widgetBox(polarization_box, "", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(self.ewp_box_8, self, "polarization_degree", "Polarization Degree [cos_s/(cos_s+sin_s)]",
                          tooltip="polarization_degree", labelWidth=310, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.ewp_box_8, self, "polarization_phase_deg", "Phase Difference [deg,0=linear,+90=ell/right]",
                          tooltip="polarization_phase_deg", labelWidth=310, valueType=float, orientation="horizontal")

        gui.rubber(self.controlArea)
        gui.rubber(self.mainArea)

        self.set_coordinates_visibility()

    def is_scanning_enabled(self):
        return True

    def set_coordinates_visibility(self):
        self.points_cartesian_box.setVisible(self.coordinates == 0)
        self.points_polar_box.setVisible(self.coordinates == 1)

    def get_lightsource(self):
        import scipy.constants as codata
        if self.units == 0:
            wavelength = codata.h * codata.c / codata.e / self.single_line_value
        else:
            wavelength = self.single_line_value * 1e-10

        if self.coordinates == 0:
            gs =  SourceGridCartesian(
                real_space_width = [self.real_space_width_x, 0.0, self.real_space_width_z],
                real_space_center=[self.real_space_center_x, 0.0, self.real_space_center_z],
                real_space_points=[self.real_space_points_x, 1, self.real_space_points_z],
                direction_space_width = [self.direction_space_width_x, self.direction_space_width_z],
                direction_space_center = [self.direction_space_center_x, self.direction_space_center_z],
                direction_space_points=[self.direction_space_points_x, self.direction_space_points_z],
                wavelength=wavelength,
                polarization_degree=self.polarization_degree,
                polarization_phase_deg=self.polarization_phase_deg,
                name = "Grid Source (Cartesian)")
        else:
            gs =  SourceGridPolar(
                real_space_width = [self.real_space_width_x, 0.0, self.real_space_width_z],
                real_space_center=[self.real_space_center_x, 0.0, self.real_space_center_z],
                real_space_points = [self.real_space_points_r, self.real_space_points_theta],
                direction_space_width = [self.direction_space_width_x, self.direction_space_width_z],
                direction_space_center=[self.direction_space_center_x, self.direction_space_center_z],
                direction_space_points = [self.direction_space_points_r, self.direction_space_points_theta],
                wavelength=wavelength,
                polarization_degree=self.polarization_degree,
                polarization_phase_deg=self.polarization_phase_deg,
                name = "Grid Source (Polar)")

        return gs

    def run_shadow4(self):

        sys.stdout = EmittingStream(textWritten=self._write_stdout)

        self._set_plot_quality()

        self.progressBarInit()

        light_source = self.get_lightsource()

        self.progressBarSet(5)

        # run shadow4

        output_beam = light_source.get_beam()

        #
        # beam plots
        #
        self._plot_results(output_beam, None, progressBarValue=80)

        #
        # script
        #
        script = light_source.to_python_code()

        script += "\n\n# test plot\nfrom srxraylib.plot.gol import plot_scatter"
        script += "\nrays = beam.get_rays()"
        script += "\nplot_scatter(1e6 * rays[:, 0], 1e6 * rays[:, 2], title='(X,Z) in microns')"

        self.shadow4_script.set_code(script)

        self.progressBarFinished()

        #
        # send beam
        #
        self.send("Shadow Data", ShadowData(beam=output_beam,
                                           number_of_rays=output_beam.get_number_of_rays(),
                                           beamline=S4Beamline(light_source=light_source)))
if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication
    a = QApplication(sys.argv)
    ow = OWGrid()
    ow.show()
    a.exec_()
    ow.saveSettings()



#! /usr/bin/env python3

# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2019 European Synchrotron Radiation Facility
# Copyright (c) 2019-2020 Lawrence Berkeley National Laboratory
# Copyright (c) 2020-now European Synchrotron Radiation Facility
# Copyright (c) 2023-now Argonne National Laboratory
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

#
# Memorandum: 
#
# Install from sources: 
#     git clone https://github.com/srio/oasys1-shadow4
#     cd oasys1-shadow4
#     python -m pip install -e . --no-deps --no-binary :all:
#
# Upload to pypi (when uploading, increment the version number):
#     python setup.py register (only once, not longer needed)
#     python setup.py sdist
#     python setup.py upload
#          
# Install from pypi:
#     pip install <name>
#

__authors__ = ["M Sanchez del Rio, Luca Rebuffi"]
__license__ = "MIT"
__date__ = "25/10/2019"

import imp
import os
import sys
from setuptools import find_packages, setup
import subprocess

NAME = 'OASYS1-shadow4'
VERSION = '0.0.14'
ISRELEASED = False

DESCRIPTION = 'oasys-shadow4: Oasys widgets for shadow4'
README_FILE = os.path.join(os.path.dirname(__file__), 'README.txt')
LONG_DESCRIPTION = open(README_FILE).read()
AUTHOR = 'M. Sanchez del Rio, L. Rebuffi'
AUTHOR_EMAIL = 'srio@esrf.eu'
URL = 'https://github.com/oasys-kit/OASYS1-SHADOW4'
DOWNLOAD_URL = 'https://github.com/oasys-kit/OASYS1-SHADOW4'
LICENSE = 'MIT'

KEYWORDS = (
    'ray tracing',
    'x-ray optics',
    'Oasys1',
    )

CLASSIFIERS = (
    'Development Status :: 4 - Beta',
    'Environment :: X11 Applications :: Qt',
    'Environment :: Console',
    'Environment :: Plugins',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Intended Audience :: Science/Research',
    )


SETUP_REQUIRES = (
                  'setuptools',
                  )

INSTALL_REQUIRES = (
                    'oasys1>=1.2.130',
                    'shadow4>=0.1.15'
                    )

PACKAGES = find_packages(exclude=('*.tests', '*.tests.*', 'tests.*', 'tests'))



PACKAGE_DATA = {
    "orangecontrib.shadow4.widgets.sources":["icons/*.png", "icons/*.jpg"],
    "orangecontrib.shadow4.widgets.optics": ["icons/*.png", "icons/*.jpg"],
    "orangecontrib.shadow4.widgets.tools":["icons/*.png", "icons/*.jpg"],
    "orangecontrib.shadow4.widgets.compatibility":["icons/*.png", "icons/*.jpg"],
    }

NAMESPACE_PACAKGES = ["orangecontrib","orangecontrib.shadow4", "orangecontrib.shadow4.widgets"]


ENTRY_POINTS = {
    'oasys.addons' : ("SHADOW4 = orangecontrib.shadow4", ),
    'oasys.widgets' : (
            "SHADOW4 Sources = orangecontrib.shadow4.widgets.sources",
            "SHADOW4 Optics = orangecontrib.shadow4.widgets.optics",
            "SHADOW4 Tools = orangecontrib.shadow4.widgets.tools",
            "SHADOW3 \u21d4 SHADOW4 = orangecontrib.shadow4.widgets.compatibility",
    ),
    }

if __name__ == '__main__':
    setup(
          name = NAME,
          version = VERSION,
          description = DESCRIPTION,
          long_description = LONG_DESCRIPTION,
          author = AUTHOR,
          author_email = AUTHOR_EMAIL,
          url = URL,
          download_url = DOWNLOAD_URL,
          license = LICENSE,
          keywords = KEYWORDS,
          classifiers = CLASSIFIERS,
          packages = PACKAGES,
          package_data = PACKAGE_DATA,
          setup_requires = SETUP_REQUIRES,
          install_requires = INSTALL_REQUIRES,
          entry_points = ENTRY_POINTS,
          namespace_packages=NAMESPACE_PACAKGES,
          include_package_data = True,
          zip_safe = False,
          )



from PyQt5.QtGui import QPalette, QColor, QFont
from PyQt5.QtCore import QRect

from orangewidget import gui
from orangewidget import widget
from orangewidget.settings import Setting

from oasys.widgets.widget import OWWidget
from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence
from oasys.widgets.gui import ConfirmDialog

from orangecontrib.shadow4.widgets.gui.ow_generic_element import GenericElement

from shadow4.sources.s4_electron_beam import S4ElectronBeam
# from syned.storage_ring.light_source import ElectronBeam
# from syned.beamline.beamline import Beamline
# from syned.util.json_tools import load_from_json_file, load_from_json_url

class OWElectronBeam(GenericElement):

    syned_file_name = Setting("Select *.json file")
    # source_name = Setting("Undefined")

    electron_energy_in_GeV = Setting(1.9)
    electron_energy_spread = Setting(0.001)
    ring_current           = Setting(0.4)
    # number_of_bunches      = Setting(400) #TODO: to be deleted in syned?

    moment_xx           = Setting(0.0)
    moment_xxp          = Setting(0.0)
    moment_xpxp         = Setting(0.0)
    moment_yy           = Setting(0.0)
    moment_yyp          = Setting(0.0)
    moment_ypyp         = Setting(0.0)

    electron_beam_size_h       = Setting(39e-6)
    electron_beam_divergence_h = Setting(31e-6)
    electron_beam_size_v       = Setting(39.2e-6)
    electron_beam_divergence_v = Setting(39.2e-6)

    electron_beam_emittance_h = Setting(0.0)
    electron_beam_emittance_v = Setting(0.0)
    electron_beam_beta_h = Setting(0.0)
    electron_beam_beta_v = Setting(0.0)
    electron_beam_alpha_h = Setting(0.0)
    electron_beam_alpha_v = Setting(0.0)
    electron_beam_eta_h = Setting(0.0)
    electron_beam_eta_v = Setting(0.0)
    electron_beam_etap_h = Setting(0.0)
    electron_beam_etap_v = Setting(0.0)

    type_of_properties = Setting(1)

    def __init__(self):
        super().__init__(show_automatic_box=False, has_footprint=False)

        self.runaction = widget.OWAction("Run Shadow4/Source", self)
        self.runaction.triggered.connect(self.run_shadow4)
        self.addAction(self.runaction)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        button = gui.button(button_box, self, "Run shadow4/source", callback=self.run_shadow4)
        font = QFont(button.font())
        font.setBold(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Blue'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)

        button = gui.button(button_box, self, "Reset Fields", callback=self.call_reset_settings)
        font = QFont(button.font())
        font.setItalic(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Red'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)
        button.setFixedWidth(150)


        self.tabs_control_area = oasysgui.tabWidget(self.controlArea)
        self.tabs_control_area.setFixedHeight(self.TABS_AREA_HEIGHT)
        self.tabs_control_area.setFixedWidth(self.CONTROL_AREA_WIDTH-5)

        self.tab_electron_beam = oasysgui.createTabPage(self.tabs_control_area, "Electron Beam Setting")

        self.electron_beam_box = oasysgui.widgetBox(self.tab_electron_beam, "Electron Beam/Machine Parameters", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.electron_beam_box, self, "electron_energy_in_GeV", "Energy [GeV]",  tooltip="electron_energy_in_GeV", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.electron_beam_box, self, "electron_energy_spread", "Energy Spread", tooltip="electron_energy_spread", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.electron_beam_box, self, "ring_current", "Ring Current [A]",        tooltip="ring_current",           labelWidth=260, valueType=float, orientation="horizontal")

        gui.comboBox(self.electron_beam_box, self, "type_of_properties", tooltip="type_of_properties", label="Electron Beam Properties", labelWidth=350,
                     items=["From 2nd Moments", "From Size/Divergence", "From Twiss parameters","Zero emittance"],
                     callback=self.set_TypeOfProperties,
                     sendSelectedValue=False, orientation="horizontal")

        self.left_box_2_1 = oasysgui.widgetBox(self.electron_beam_box, "", addSpace=False, orientation="vertical", height=150)

        oasysgui.lineEdit(self.left_box_2_1, self, "moment_xx",   "<x x>   [m^2]",   tooltip="moment_xx",   labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_1, self, "moment_xxp",  "<x x'>  [m.rad]", tooltip="moment_xxp",  labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_1, self, "moment_xpxp", "<x' x'> [rad^2]", tooltip="moment_xpxp", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_1, self, "moment_yy",   "<y y>   [m^2]",   tooltip="moment_yy",   labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_1, self, "moment_yyp",  "<y y'>  [m.rad]", tooltip="moment_yyp",  labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_1, self, "moment_ypyp", "<y' y'> [rad^2]", tooltip="moment_ypyp", labelWidth=260, valueType=float, orientation="horizontal")


        self.left_box_2_2 = oasysgui.widgetBox(self.electron_beam_box, "", addSpace=False, orientation="vertical", height=150)

        oasysgui.lineEdit(self.left_box_2_2, self, "electron_beam_size_h",       "Horizontal Beam Size \u03c3x [m]",          tooltip="electron_beam_size_h", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_2, self, "electron_beam_size_v",       "Vertical Beam Size \u03c3y [m]",            tooltip="electron_beam_size_v",  labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_2, self, "electron_beam_divergence_h", "Horizontal Beam Divergence \u03c3'x [rad]", tooltip="electron_beam_divergence_h", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_2, self, "electron_beam_divergence_v", "Vertical Beam Divergence \u03c3'y [rad]",   tooltip="electron_beam_divergence_v", labelWidth=260, valueType=float, orientation="horizontal")

        self.left_box_2_3 = oasysgui.widgetBox(self.electron_beam_box, "", addSpace=False, orientation="horizontal",height=150)
        self.left_box_2_3_l = oasysgui.widgetBox(self.left_box_2_3, "", addSpace=False, orientation="vertical")
        self.left_box_2_3_r = oasysgui.widgetBox(self.left_box_2_3, "", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(self.left_box_2_3_l, self, "electron_beam_emittance_h", "\u03B5x [m.rad]",tooltip="electron_beam_emittance_h",labelWidth=75, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_3_l, self, "electron_beam_alpha_h",     "\u03B1x",        tooltip="electron_beam_alpha_h",    labelWidth=75, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_3_l, self, "electron_beam_beta_h",      "\u03B2x [m]",    tooltip="electron_beam_beta_h",     labelWidth=75, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_3_l, self, "electron_beam_eta_h",       "\u03B7x",        tooltip="electron_beam_eta_h",      labelWidth=75, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_3_l, self, "electron_beam_etap_h",      "\u03B7'x",       tooltip="electron_beam_etap_h",     labelWidth=75, valueType=float, orientation="horizontal")


        oasysgui.lineEdit(self.left_box_2_3_r, self, "electron_beam_emittance_v", "\u03B5y [m.rad]",tooltip="",labelWidth=75, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_3_r, self, "electron_beam_alpha_v",     "\u03B1y",        tooltip="",labelWidth=75,valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_3_r, self, "electron_beam_beta_v",      "\u03B2y [m]",    tooltip="",labelWidth=75, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_3_r, self, "electron_beam_eta_v",       "\u03B7y",        tooltip="",labelWidth=75, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_3_r, self, "electron_beam_etap_v",      "\u03B7'y",       tooltip="",labelWidth=75, valueType=float, orientation="horizontal")

        self.set_TypeOfProperties()

        gui.rubber(self.controlArea)


    def set_TypeOfProperties(self):
        self.left_box_2_1.setVisible(self.type_of_properties == 0)
        self.left_box_2_2.setVisible(self.type_of_properties == 1)
        self.left_box_2_3.setVisible(self.type_of_properties == 2)


    def check_data(self):
        congruence.checkStrictlyPositiveNumber(self.electron_energy_in_GeV , "Energy")
        congruence.checkStrictlyPositiveNumber(self.electron_energy_spread, "Energy Spread")
        congruence.checkStrictlyPositiveNumber(self.ring_current, "Ring Current")

        if self.type_of_properties == 0:
            congruence.checkPositiveNumber(self.moment_xx   , "Moment xx")
            congruence.checkPositiveNumber(self.moment_xpxp , "Moment xpxp")
            congruence.checkPositiveNumber(self.moment_yy   , "Moment yy")
            congruence.checkPositiveNumber(self.moment_ypyp , "Moment ypyp")
        elif self.type_of_properties == 1:
            congruence.checkPositiveNumber(self.electron_beam_size_h       , "Horizontal Beam Size")
            congruence.checkPositiveNumber(self.electron_beam_divergence_h , "Vertical Beam Size")
            congruence.checkPositiveNumber(self.electron_beam_size_v       , "Horizontal Beam Divergence")
            congruence.checkPositiveNumber(self.electron_beam_divergence_v , "Vertical Beam Divergence")
        elif self.type_of_properties == 2:
            congruence.checkPositiveNumber(self.electron_beam_emittance_h, "Horizontal Beam Emittance")
            congruence.checkPositiveNumber(self.electron_beam_emittance_v, "Vertical Beam Emittance")
            congruence.checkNumber(self.electron_beam_alpha_h, "Horizontal Beam Alpha")
            congruence.checkNumber(self.electron_beam_alpha_v, "Vertical Beam Alpha")
            congruence.checkNumber(self.electron_beam_beta_h, "Horizontal Beam Beta")
            congruence.checkNumber(self.electron_beam_beta_v, "Vertical Beam Beta")
            congruence.checkNumber(self.electron_beam_eta_h, "Horizontal Beam Dispersion Eta")
            congruence.checkNumber(self.electron_beam_eta_v, "Vertical Beam Dispersion Eta")
            congruence.checkNumber(self.electron_beam_etap_h, "Horizontal Beam Dispersion Eta'")
            congruence.checkNumber(self.electron_beam_etap_v, "Vertical Beam Dispersion Eta'")

        self.check_magnetic_structure()


    def run_shadow4(self):
        raise Exception("To be defined in the superclass")

    def get_electron_beam(self): #to do: remove script
        electron_beam = S4ElectronBeam(energy_in_GeV=self.electron_energy_in_GeV,
                                     energy_spread=self.electron_energy_spread,
                                     current=self.ring_current,
                                     # number_of_bunches=self.number_of_bunches,
                                     )

        script = "\n# electron beam"
        script += "\nfrom shadow4.sources.s4_electron_beam import S4ElectronBeam"
        script += "\nelectron_beam = S4ElectronBeam(energy_in_GeV=%g,energy_spread=%g,current=%g)" % \
                  (self.electron_energy_in_GeV,self.electron_energy_spread,self.ring_current)

        recalculate_fields = False

        if self.type_of_properties == 0:
            electron_beam.set_moments_horizontal(self.moment_xx,self.moment_xxp,self.moment_xpxp)
            electron_beam.set_moments_vertical(self.moment_yy, self.moment_yyp, self.moment_ypyp)

            script += "\nelectron_beam.set_moments_horizontal(%g,%g,%g)" % (self.moment_xx,self.moment_xxp,self.moment_xpxp)
            script +=   "\nelectron_beam.set_moments_vertical(%g,%g,%g)" % (self.moment_yy, self.moment_yyp, self.moment_ypyp)

            if recalculate_fields:

                x, xp, y, yp = electron_beam.get_sigmas_all()

                self.electron_beam_size_h = x
                self.electron_beam_size_v = y
                self.electron_beam_divergence_h = xp
                self.electron_beam_divergence_v = yp

                twiss_all = electron_beam.get_twiss_no_dispersion_all()
                self.electron_beam_emittance_h = twiss_all[0]
                self.electron_beam_alpha_h     = twiss_all[1]
                self.electron_beam_beta_h      = twiss_all[2]
                self.electron_beam_eta_h       = 0.0
                self.electron_beam_etap_h      = 0.0
                self.electron_beam_emittance_v = twiss_all[3]
                self.electron_beam_alpha_v     = twiss_all[4]
                self.electron_beam_beta_v      = twiss_all[5]
                self.electron_beam_eta_v       = 0.0
                self.electron_beam_etap_v      = 0.0


        elif self.type_of_properties == 1:
            electron_beam.set_sigmas_all(sigma_x=self.electron_beam_size_h,
                                         sigma_y=self.electron_beam_size_v,
                                         sigma_xp=self.electron_beam_divergence_h,
                                         sigma_yp=self.electron_beam_divergence_v)
            script += "\nelectron_beam.set_sigmas_all(sigma_x=%g,sigma_y=%g,sigma_xp=%g,sigma_yp=%g)" % \
                        (self.electron_beam_size_h,
                        self.electron_beam_size_v,
                        self.electron_beam_divergence_h,
                        self.electron_beam_divergence_v)
            if recalculate_fields:
                moments_all = electron_beam.get_moments_all()

                self.moment_xx   = moments_all[0]
                self.moment_xxp  = moments_all[1]
                self.moment_xpxp = moments_all[2]
                self.moment_yy   = moments_all[3]
                self.moment_yy   = moments_all[4]
                self.moment_ypyp = moments_all[5]

                twiss_all = electron_beam.get_twiss_no_dispersion_all()
                self.electron_beam_emittance_h = twiss_all[0]
                self.electron_beam_alpha_h     = twiss_all[1]
                self.electron_beam_beta_h      = twiss_all[2]
                self.electron_beam_eta_h       = 0.0
                self.electron_beam_etap_h      = 0.0
                self.electron_beam_emittance_v = twiss_all[3]
                self.electron_beam_alpha_v     = twiss_all[4]
                self.electron_beam_beta_v      = twiss_all[5]
                self.electron_beam_eta_v       = 0.0
                self.electron_beam_etap_v      = 0.0

        elif self.type_of_properties == 2:
            electron_beam.set_twiss_horizontal(self.electron_beam_emittance_h,
                                             self.electron_beam_alpha_h,
                                             self.electron_beam_beta_h,
                                             self.electron_beam_eta_h,
                                             self.electron_beam_etap_h)
            electron_beam.set_twiss_vertical(self.electron_beam_emittance_v,
                                             self.electron_beam_alpha_v,
                                             self.electron_beam_beta_v,
                                             self.electron_beam_eta_v,
                                             self.electron_beam_etap_v)
            script += "\nelectron_beam.set_twiss_horizontal(%g,%g,%g,%g,%g)" % (
                  self.electron_beam_emittance_h,
                  self.electron_beam_alpha_h,
                  self.electron_beam_beta_h,
                  self.electron_beam_eta_h,
                  self.electron_beam_etap_h,
            )
            script += "\nelectron_beam.set_twiss_vertical(%g,%g,%g,%g,%g)" % (
                  self.electron_beam_emittance_v,
                  self.electron_beam_alpha_v,
                  self.electron_beam_beta_v,
                  self.electron_beam_eta_v,
                  self.electron_beam_etap_v,
            )



            if recalculate_fields:
                x, xp, y, yp = electron_beam.get_sigmas_all()

                self.electron_beam_size_h = x
                self.electron_beam_size_v = y
                self.electron_beam_divergence_h = xp
                self.electron_beam_divergence_v = yp

                moments_all = electron_beam.get_moments_all()

                self.moment_xx   = moments_all[0]
                self.moment_xxp  = moments_all[1]
                self.moment_xpxp = moments_all[2]
                self.moment_yy   = moments_all[3]
                self.moment_yy   = moments_all[4]
                self.moment_ypyp = moments_all[5]
        elif self.type_of_properties == 3:
            electron_beam.set_moments_all(0,0,0,0,0,0)
            script += "\nelectron_beam.set_moments_all(0,0,0,0,0,0)"

        return electron_beam

    def populate_fields_from_electron_beam(self, electron_beam):

        self.electron_energy_in_GeV = electron_beam.energy()
        self.electron_energy_spread = electron_beam._energy_spread
        self.ring_current = electron_beam.current()

        moment_xx, moment_xxp, moment_xpxp, moment_yy, moment_yyp, moment_ypyp = electron_beam.get_moments_all()
        self.moment_xx   = moment_xx
        self.moment_xxp  = moment_xxp
        self.moment_xpxp = moment_xpxp
        self.moment_yy   = moment_yy
        self.moment_yyp  = moment_yyp
        self.moment_ypyp = moment_ypyp

        x, xp, y, yp = electron_beam.get_sigmas_all()
        self.electron_beam_size_h       = x
        self.electron_beam_size_v       = y
        self.electron_beam_divergence_h = xp
        self.electron_beam_divergence_v = yp

        self.type_of_properties = 1
        self.set_TypeOfProperties()



import numpy
from orangewidget import gui
from orangewidget.settings import Setting

from oasys.widgets import gui as oasysgui

from syned.beamline.element_coordinates import ElementCoordinates

from shadow4.beamline.optical_elements.ideal_elements.s4_ideal_lens import S4IdealLensElement, S4SuperIdealLensElement, S4IdealLens, S4SuperIdealLens

from orangecontrib.shadow4.widgets.gui.ow_optical_element_with_surface_shape import OWOpticalElementWithSurfaceShape


class OWIdealLens(OWOpticalElementWithSurfaceShape):
    name        = "Ideal Lens"
    description = "Shadow Ideal Lens"
    icon        = "icons/ideal_lens.png"

    priority = 3.2

    focal_x    = Setting(0.0)
    focal_z    = Setting(0.0)

    focal_p_x    = Setting(0.0)
    focal_p_z    = Setting(0.0)
    focal_q_x    = Setting(0.0)
    focal_q_z    = Setting(0.0)

    ideal_lens_type = Setting(0)

    def __init__(self):
        super().__init__(has_footprint=False)

    def populate_tab_position(self, tab_position):
        self.orientation_box = oasysgui.widgetBox(tab_position, "Optical Element Orientation", addSpace=True, orientation="vertical")

        oasysgui.lineEdit(self.orientation_box, self, "source_plane_distance", "Source Plane Distance [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="source_plane_distance")
        oasysgui.lineEdit(self.orientation_box, self, "image_plane_distance", "Image Plane Distance [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="image_plane_distance")

    def create_basic_settings_subtabs(self, tabs_basic_settings):
        return oasysgui.createTabPage(tabs_basic_settings, "Ideal Lens Parameters")  # to be populated

    def populate_basic_setting_subtabs(self, basic_setting_subtabs):
        tab_ideal_lens = basic_setting_subtabs

        box_ideal_lens = oasysgui.widgetBox(tab_ideal_lens, "", addSpace=False, orientation="vertical", height=180)

        gui.comboBox(box_ideal_lens, self, "ideal_lens_type", label="Ideal Lens Type", labelWidth=350,
                     items=["Simple", "Super"],
                     callback=self.set_ideal_lens_type, sendSelectedValue=False, orientation="horizontal",
                     tooltip="ideal_lens_type")

        gui.separator(box_ideal_lens)

        self.box_focal_distances = oasysgui.widgetBox(box_ideal_lens, "Focal Distances", addSpace=False, orientation="vertical", height=140)
        self.box_p_q_distances   = oasysgui.widgetBox(box_ideal_lens, "p,q Distances",   addSpace=False, orientation="vertical", height=140)

        oasysgui.lineEdit(self.box_focal_distances, self, "focal_x", "focal distance X [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="focal_x")
        oasysgui.lineEdit(self.box_focal_distances, self, "focal_z", "focal distance Z [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="focal_z")

        oasysgui.lineEdit(self.box_p_q_distances, self, "focal_p_x", "p(X) [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="focal_p_x")
        oasysgui.lineEdit(self.box_p_q_distances, self, "focal_p_z", "p(Z) [m]",
                          labelWidth=260, valueType=float, orientation="horizontal", tooltip="focal_p_z")
        oasysgui.lineEdit(self.box_p_q_distances, self, "focal_q_x", "q(X) [m]",
                          labelWidth=260, valueType=float, orientation="horizontal", tooltip="focal_q_x")
        oasysgui.lineEdit(self.box_p_q_distances, self, "focal_q_z", "q(Z) [m]",
                          labelWidth=260, valueType=float, orientation="horizontal", tooltip="focal_q_z")

        self.set_ideal_lens_type()

    def set_ideal_lens_type(self):
        self.box_focal_distances.setVisible(self.ideal_lens_type == 0)
        self.box_p_q_distances.setVisible(self.ideal_lens_type == 1)

    # ----------------------------------------------------
    # from OpticalElement

    def get_coordinates_instance(self):
        return ElementCoordinates(
                p=self.source_plane_distance,
                q=self.image_plane_distance,
                angle_radial=0.0,
                angle_azimuthal=0.0,
                angle_radial_out=numpy.pi,
                )

    def get_optical_element_instance(self):
        try:    name = self.getNode().title
        except: name = "Ideal Lens"

        if self.ideal_lens_type == 0:   return S4IdealLens(name=name,
                                                           focal_x=self.focal_x,
                                                           focal_y=self.focal_z)
        elif self.ideal_lens_type == 1: return S4SuperIdealLens(name=name,
                                                                focal_p_x=self.focal_p_x,
                                                                focal_p_y=self.focal_p_z,
                                                                focal_q_x=self.focal_q_x,
                                                                focal_q_y=self.focal_q_z)

    def get_beamline_element_instance(self):
        if self.ideal_lens_type == 0:   return S4IdealLensElement()
        elif self.ideal_lens_type == 1: return S4SuperIdealLensElement()


if __name__ == "__main__":
    from shadow4.beamline.s4_beamline import S4Beamline
    from orangecontrib.shadow4.util.shadow4_objects import ShadowData
    import sys
    def get_test_beam():
        from shadow4.sources.source_geometrical.source_geometrical import SourceGeometrical
        light_source = SourceGeometrical(name='SourceGeometrical', nrays=5000, seed=5676561)
        light_source.set_spatial_type_rectangle(width=0.001000, height=0.001000)
        light_source.set_angular_distribution_flat(hdiv1=0.000000, hdiv2=0.000000, vdiv1=0.000000, vdiv2=0.000000)
        light_source.set_energy_distribution_singleline(5000.000000, unit='A')
        light_source.set_polarization(polarization_degree=1.000000, phase_diff=0.000000, coherent_beam=0)
        beam = light_source.get_beam()

        return ShadowData(beam=beam, beamline=S4Beamline(light_source=light_source))

    from PyQt5.QtWidgets import QApplication
    a = QApplication(sys.argv)
    ow = OWIdealLens()
    ow.set_shadow_data(get_test_beam())

    ow.source_plane_distance = 0
    ow.image_plane_distance = 5000.0
    ow.focal_x = 5000.0
    ow.focal_z = 5000.0
    ow.run_shadow4()

    ow.show()
    a.exec_()
    ow.saveSettings()




from oasys.widgets import gui as oasysgui


def plot_data1D(x, y,
                title="", xtitle="", ytitle="",
                log_x=False, log_y=False, color='blue', replace=True, control=False,
                xrange=None, yrange=None, symbol=''):



    plot_widget_id = oasysgui.plotWindow(parent=None,
                                         backend=None,
                                         resetzoom=True,
                                         autoScale=False,
                                         logScale=True,
                                         grid=True,
                                         curveStyle=True,
                                         colormap=False,
                                         aspectRatio=False,
                                         yInverted=False,
                                         copy=True,
                                         save=True,
                                         print_=True,
                                         control=control,
                                         position=True,
                                         roi=False,
                                         mask=False,
                                         fit=False)


    plot_widget_id.setDefaultPlotLines(True)
    plot_widget_id.setActiveCurveColor(color='blue')
    plot_widget_id.setGraphXLabel(xtitle)
    plot_widget_id.setGraphYLabel(ytitle)


    plot_widget_id.addCurve(x, y, title, symbol=symbol, color=color, xlabel=xtitle, ylabel=ytitle, replace=replace)  # '+', '^', ','

    if not xtitle is None: plot_widget_id.setGraphXLabel(xtitle)
    if not ytitle is None: plot_widget_id.setGraphYLabel(ytitle)
    if not title is None:  plot_widget_id.setGraphTitle(title)

    plot_widget_id.resetZoom()
    plot_widget_id.replot()
    plot_widget_id.setActiveCurve(title)


    plot_widget_id.setXAxisLogarithmic(log_x)
    plot_widget_id.setYAxisLogarithmic(log_y)


    if xrange is not None:
        plot_widget_id.setGraphXLimits(xrange[0] ,xrange[1])
    if yrange is not None:
        plot_widget_id.setGraphYLimits(yrange[0] ,yrange[1])

    if min(y) < 0:
        if log_y:
            plot_widget_id.setGraphYLimits(min(y ) *1.2, max(y ) *1.2)
        else:
            plot_widget_id.setGraphYLimits(min(y ) *1.01, max(y ) *1.01)
    else:
        if log_y:
            plot_widget_id.setGraphYLimits(min(y), max(y ) *1.2)
        else:
            plot_widget_id.setGraphYLimits(min(y ) *0.99, max(y ) *1.01)

    return plot_widget_id




import numpy
import sys

from PyQt5.QtGui import QPalette, QColor, QFont


from orangewidget import widget
from orangewidget import gui
from orangewidget.settings import Setting

from oasys.widgets import gui as oasysgui
from oasys.util.oasys_util import EmittingStream

from orangecontrib.shadow4.widgets.gui.ow_generic_element import GenericElement

from syned.widget.widget_decorator import WidgetDecorator

from shadow4.beamline.optical_elements.ideal_elements.s4_beam_movement import S4BeamMovement, S4BeamMovementElement

from orangecontrib.shadow4.util.shadow4_objects import ShadowData
from orangecontrib.shadow4.util.shadow4_util import ShadowCongruence


class OWBeamMovement(GenericElement, WidgetDecorator):

    name = "Beam movements"
    description = "Shadow Beam Movement"
    icon = "icons/beam_movement.png"
    priority = 10.1

    inputs = [("Shadow Data", ShadowData, "set_shadow_data")]
    WidgetDecorator.append_syned_input_data(inputs)

    outputs = [{"name":"Shadow Data",
                "type":ShadowData,
                "doc":"",}]

    #########################################################
    # Position
    #########################################################
    apply_flag = Setting(1)
    translation_x  = Setting(0.0)
    translation_y  = Setting(0.0)
    translation_z  = Setting(0.0)
    rotation_x     = Setting(0.0)
    rotation_y     = Setting(0.0)
    rotation_z     = Setting(0.0)

    input_data = None
    beamline = None


    def __init__(self):
        super().__init__(show_automatic_box=True, has_footprint=False)

        #
        # main buttons
        #
        self.runaction = widget.OWAction("Run Shadow4/Trace", self)
        self.runaction.triggered.connect(self.run_shadow4)
        self.addAction(self.runaction)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        button = gui.button(button_box, self, "Run shadow4/trace", callback=self.run_shadow4)
        font = QFont(button.font())
        font.setBold(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Blue'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)

        button = gui.button(button_box, self, "Reset Fields", callback=self.call_reset_settings)
        font = QFont(button.font())
        font.setItalic(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Red'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)
        button.setFixedWidth(150)

        #
        # tabs
        #
        self.tabs_control_area = oasysgui.tabWidget(self.controlArea)
        self.tabs_control_area.setFixedHeight(self.TABS_AREA_HEIGHT)
        self.tabs_control_area.setFixedWidth(self.CONTROL_AREA_WIDTH-5)


        self.tab_movement = oasysgui.createTabPage(self.tabs_control_area, "Movement")           # to be populated

        #
        # populate tabs with widgets
        #

        #########################################################
        # Position
        #########################################################
        self.populate_tab_movement()

        self.update_panels()
        #
        #
        #

        gui.rubber(self.controlArea)
        gui.rubber(self.mainArea)

    def resetSettings(self):
        self.translation_x = 0.0
        self.translation_y = 0.0
        self.translation_z = 0.0
        self.rotation_x = 0.0
        self.rotation_y = 0.0
        self.rotation_z = 0.0


    def populate_tab_movement(self):

        gui.comboBox(self.tab_movement, self, "apply_flag", label="Apply movements",
                     labelWidth=250, items=["Off", "On"], callback=self.update_panels,
                     sendSelectedValue=False, orientation="horizontal", tooltip="angles_respect_to")


        self.translation_box = oasysgui.widgetBox(self.tab_movement, "Translation", addSpace=True, orientation="vertical")
        self.rotation_box = oasysgui.widgetBox(self.tab_movement, "Rotation", addSpace=True, orientation="vertical")


        oasysgui.lineEdit(self.translation_box, self, "translation_x", "Translation along X [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="translation_x")
        oasysgui.lineEdit(self.translation_box, self, "translation_y", "Translation along Y [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="translation_y")
        oasysgui.lineEdit(self.translation_box, self, "translation_z", "Translation along Z [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="translation_z")

        oasysgui.lineEdit(self.rotation_box, self, "rotation_x", "1) Rotation along X [rad]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="rotation_x")
        oasysgui.lineEdit(self.rotation_box, self, "rotation_y", "2) Rotation along Y [rad]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="rotation_y")
        oasysgui.lineEdit(self.rotation_box, self, "rotation_z", "3) Rotation along Z [rad]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="rotation_z")


    def update_panels(self):

        if self.apply_flag:
            self.translation_box.setVisible(True)
            self.rotation_box.setVisible(True)
        else:
            self.translation_box.setVisible(False)
            self.rotation_box.setVisible(False)


    def get_oe_instance(self):
        return S4BeamMovement(
            apply_flag = self.apply_flag,
            translation_x = self.translation_x,
            translation_y = self.translation_y,
            translation_z = self.translation_z,
            rotation_x = self.rotation_x,
            rotation_y = self.rotation_y,
            rotation_z = self.rotation_z,
        )

    def get_element_instance(self):
        optical_element = S4BeamMovementElement()
        optical_element.set_optical_element(self.get_oe_instance())
        # optical_element.set_coordinates()
        optical_element.set_input_beam(self.input_data.beam)
        return optical_element

    def set_shadow_data(self, input_data):
        self.not_interactive = self._check_not_interactive_conditions(input_data)

        self._on_receiving_input()

        if ShadowCongruence.check_empty_data(input_data):
            self.input_data = input_data.duplicate()
            if self.is_automatic_run: self.run_shadow4()


    def run_shadow4(self):
        self.shadow_output.setText("")

        sys.stdout = EmittingStream(textWritten=self._write_stdout)

        element = self.get_element_instance()
        print(element.info())

        self.progressBarInit()

        output_beam, _ = element.trace_beam()

        beamline = self.input_data.beamline.duplicate()
        beamline.append_beamline_element(element)

        self._set_plot_quality()

        self._plot_results(output_beam, None, progressBarValue=80)

        self.progressBarFinished()

        #
        # script
        #
        script = beamline.to_python_code()
        script += "\n\n\n# test plot"
        script += "\nif True:"
        script += "\n   from srxraylib.plot.gol import plot_scatter"
        script += "\n   # plot_scatter(beam.get_photon_energy_eV(nolost=1), beam.get_column(23, nolost=1), title='(Intensity,Photon Energy)', plot_histograms=0)"
        script += "\n   plot_scatter(1e6 * beam.get_column(1, nolost=1), 1e6 * beam.get_column(3, nolost=1), title='(X,Z) in microns')"
        self.shadow4_script.set_code(script)

        #
        # send beam
        #
        self.send("Shadow Data", ShadowData(beam=output_beam, beamline=beamline))


if __name__ == "__main__":
    from shadow4.beamline.s4_beamline import S4Beamline
    def get_test_beam():
        # electron beam
        from syned.storage_ring.light_source import ElectronBeam
        electron_beam = ElectronBeam(energy_in_GeV=6, energy_spread=0.001, current=0.2)
        electron_beam.set_sigmas_all(sigma_x=3.01836e-05, sigma_y=3.63641e-06, sigma_xp=4.36821e-06,
                                     sigma_yp=1.37498e-06)

        # Gaussian undulator
        from shadow4.sources.undulator.s4_undulator import S4Undulator
        undulator_source = S4Undulator(
            period_length=0.0159999,
            number_of_periods=100,
            emin=2700.136,
            emax=2700.136,
            flag_emittance=1,  # Use emittance (0=No, 1=Yes)
        )
        undulator_source.set_energy_monochromatic(2700.14)

        from shadow4.sources.undulator.s4_undulator_light_source import S4UndulatorLightSource
        light_source = S4UndulatorLightSource(name='GaussianUndulator', electron_beam=electron_beam,
                                             magnetic_structure=undulator_source)

        beam = light_source.get_beam_in_gaussian_approximation()

        return ShadowData(beam=beam, beamline=S4Beamline(light_source=light_source))

    from PyQt5.QtWidgets import QApplication
    a = QApplication(sys.argv)
    ow = OWBeamMovement()
    ow.set_shadow_data(get_test_beam())
    ow.show()
    a.exec_()
    ow.saveSettings()
